from __future__ import annotations
import numpy as np
from silx.gui.plot3d.items.core import ItemChangedType
from silx.gui.plot3d.items import DataItem3D, ColormapMixIn
from silx.gui.plot3d.scene import primitives
from silx.io.url import DataUrl
from ewoks3dxrd.nexus.grains import read_grains
from ewoks3dxrd.nexus.utils import group_exists


class Spheres(DataItem3D, ColormapMixIn):

    def __init__(self, positions=None, radii=None, values=None, parent=None):
        DataItem3D.__init__(self, parent=parent)
        ColormapMixIn.__init__(self)
        self._colorVal = values
        self._radiiNorm = 0.0001
        if positions is None:
            positions = np.zeros((0, 3), dtype=np.float32)
        if radii is None:
            radii = np.zeros((0, 1), dtype=np.float32)
        self._sphere = primitives.Spheres(positions, self._radiiNorm * radii)
        self._getScenePrimitive().children.append(self._sphere)
        self._colormapChanged()

    def _colormapChanged(self):
        super()._colormapChanged()
        self.setValues(self._colorVal, copy=False)

    def setRadii(self, radii: np.array):
        position = self._sphere.getAttribute(name="position", copy=False)
        if len(position) != len(radii):
            raise ValueError("Shape mismatch")
        self._sphere.setAttribute(name="radius", array=self._radiiNorm * radii)

    def setPosition(self, newPosition: np.array):
        oldPosition = self._sphere.getAttribute(name="position", copy=False)
        if newPosition.size != oldPosition.size:
            raise ValueError("Shape mismatch")
        self._sphere.setAttribute(name="position", array=newPosition)

    def setValues(self, values: np.array, copy: bool = True):
        self._colorVal = values
        if self._colorVal is None:
            self._sphere.setAttribute(name="color", array=(1.0, 1.0, 1.0, 1.0))
            return

        colors = self.getColormap().applyToData(self._colorVal) / 255
        self._sphere.setAttribute(name="color", array=colors, copy=True)

    def setRadiiNorm(self, radiiNorm: float):
        radii = self._sphere.getAttribute(name="radius") / self._radiiNorm
        self._radiiNorm = radiiNorm
        self.setRadii(radii=radii)

    def setData(self, positions: np.array, radii: np.array, values: np.array):
        self._sphere.setAttribute(name="position", array=positions)
        self._sphere.setAttribute(name="radius", array=self._radiiNorm * radii)
        self._colorVal = values
        colors = self.getColormap().applyToData(self._colorVal) / 255
        self._sphere.setAttribute(name="color", array=colors)
        self._updated(ItemChangedType.DATA)


def build_grain_spheres(grain_url: str, sphere_scale: float = 0.0001) -> Spheres:
    data_url = DataUrl(grain_url)
    group_exists(filename=data_url.file_path(), data_group_path=data_url.data_path())
    grains = read_grains(
        grain_file_h5=data_url.file_path(),
        entry_name=data_url.data_path(),
        process_group_name="",
    )

    position = np.transpose(np.array([grain.translation for grain in grains])).reshape(
        -1, 3
    )

    sphere = Spheres(
        positions=position,
        radii=np.array([float(grain.mean_intensity) for grain in grains]),
        values=np.array([float(grain.npks) for grain in grains]),
    )
    sphere.setRadiiNorm(radiiNorm=sphere_scale)
    return sphere



from __future__ import annotations

import os

from silx.gui import qt

from ..common.collapsible_widget import CollapsibleWidget


class GeometryParameterGroupBox(CollapsibleWidget):

    def __init__(self, parent: qt.QWidget | None = None, **kwargs) -> None:
        super().__init__("Edit parameters", parent=parent, **kwargs)
        layout = qt.QFormLayout()
        self.setCollapsed(True)

        self._chi = qt.QLineEdit()
        self._distance = qt.QLineEdit()
        self._fit_tolerance = qt.QLineEdit()
        self._min_bin_prob = qt.QLineEdit()
        self._no_bins = qt.QLineEdit()
        self._o11 = qt.QLineEdit()
        self._o12 = qt.QLineEdit()
        self._o21 = qt.QLineEdit()
        self._o22 = qt.QLineEdit()
        self._omegasign = qt.QLineEdit()
        self._t_x = qt.QLineEdit()
        self._t_y = qt.QLineEdit()
        self._t_z = qt.QLineEdit()
        self._tilt_x = qt.QLineEdit()
        self._tilt_y = qt.QLineEdit()
        self._tilt_z = qt.QLineEdit()
        self._wavelength = qt.QLineEdit()
        self._wedge = qt.QLineEdit()
        self._weight_hist_intensities = qt.QLineEdit()
        self._y_center = qt.QLineEdit()
        self._y_size = qt.QLineEdit()
        self._z_center = qt.QLineEdit()
        self._z_size = qt.QLineEdit()

        layout.addRow("chi", self._chi)
        layout.addRow("distance", self._distance)
        layout.addRow("fit_tolerance", self._fit_tolerance)
        layout.addRow("min_bin_prob", self._min_bin_prob)
        layout.addRow("no_bins", self._no_bins)
        layout.addRow("o11", self._o11)
        layout.addRow("o12", self._o12)
        layout.addRow("o21", self._o21)
        layout.addRow("o22", self._o22)
        layout.addRow("omegasign", self._omegasign)
        layout.addRow("t_x", self._t_x)
        layout.addRow("t_y", self._t_y)
        layout.addRow("t_z", self._t_z)
        layout.addRow("tilt_x", self._tilt_x)
        layout.addRow("tilt_y", self._tilt_y)
        layout.addRow("tilt_z", self._tilt_z)
        layout.addRow("wavelength", self._wavelength)
        layout.addRow("wedge", self._wedge)
        layout.addRow("weight_hist_intensities", self._weight_hist_intensities)
        layout.addRow("y_center", self._y_center)
        layout.addRow("y_size", self._y_size)
        layout.addRow("z_center", self._z_center)
        layout.addRow("z_size", self._z_size)
        self.setLayout(layout)

    def getGeometryParameters(self) -> dict[str, str]:
        key_to_widget = {
            "chi": self._chi,
            "distance": self._distance,
            "fit_tolerance": self._fit_tolerance,
            "min_bin_prob": self._min_bin_prob,
            "no_bins": self._no_bins,
            "o11": self._o11,
            "o12": self._o12,
            "o21": self._o21,
            "o22": self._o22,
            "omegasign": self._omegasign,
            "t_x": self._t_x,
            "t_y": self._t_y,
            "t_z": self._t_z,
            "tilt_x": self._tilt_x,
            "tilt_y": self._tilt_y,
            "tilt_z": self._tilt_z,
            "wavelength": self._wavelength,
            "wedge": self._wedge,
            "weight_hist_intensities": self._weight_hist_intensities,
            "y_center": self._y_center,
            "y_size": self._y_size,
            "z_center": self._z_center,
            "z_size": self._z_size,
        }

        parameters = {
            key: widget.text().strip()
            for key, widget in key_to_widget.items()
            if widget.text().strip() != ""
        }

        return parameters

    def fillGeometryValues(self, filePath: str):
        if not os.path.exists(filePath):
            qt.QMessageBox.critical(
                self, "File Not Found", f"File does not exist:\n{filePath}"
            )
            return

        if not filePath.lower().endswith(".par"):
            qt.QMessageBox.warning(
                self, "Invalid File", "Selected file must end with `.par`."
            )
            return
        self._parse_par_file(filePath)

    def _parse_par_file(self, filePath: str):
        key_to_widget = {
            "chi": self._chi,
            "distance": self._distance,
            "fit_tolerance": self._fit_tolerance,
            "min_bin_prob": self._min_bin_prob,
            "no_bins": self._no_bins,
            "o11": self._o11,
            "o12": self._o12,
            "o21": self._o21,
            "o22": self._o22,
            "omegasign": self._omegasign,
            "t_x": self._t_x,
            "t_y": self._t_y,
            "t_z": self._t_z,
            "tilt_x": self._tilt_x,
            "tilt_y": self._tilt_y,
            "tilt_z": self._tilt_z,
            "wavelength": self._wavelength,
            "wedge": self._wedge,
            "weight_hist_intensities": self._weight_hist_intensities,
            "y_center": self._y_center,
            "y_size": self._y_size,
            "z_center": self._z_center,
            "z_size": self._z_size,
        }

        with open(filePath, "r") as f:
            for line in f:
                line = line.strip()
                if not line:
                    continue
                key, value = map(str.strip, line.split(" ", 1))
                if key in key_to_widget:
                    key_to_widget[key].setText(value)



from silx.gui import qt


class DoubleSpinBoxRangeWidget(qt.QWidget):
    """A widget with two QDoubleSpinBoxes representing a range (min, max) in a single row."""

    def __init__(
        self,
        parent=None,
        minimum=0.0,
        maximum=180.0,
        decimals=2,
        step=0.1,
        default=(0.0, 180.0),
    ):
        super().__init__(parent)
        layout = qt.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        self._minSpin = qt.QDoubleSpinBox(self)
        self._maxSpin = qt.QDoubleSpinBox(self)

        for spin in (self._minSpin, self._maxSpin):
            spin.setDecimals(decimals)
            spin.setRange(minimum, maximum)
            spin.setSingleStep(step)

        self._minSpin.setValue(default[0])
        self._maxSpin.setValue(default[1])

        layout.addWidget(self._minSpin)
        layout.addWidget(qt.QLabel("to"))
        layout.addWidget(self._maxSpin)

    def getValue(self) -> tuple[float, float]:
        return (self._minSpin.value(), self._maxSpin.value())

    def setValue(self, values: tuple[float, float]):
        self._minSpin.setValue(values[0])
        self._maxSpin.setValue(values[1])



from __future__ import annotations

from typing import Optional

import numpy as np
from ewoksorange.gui.orange_imports import Input
from ImageD11.unitcell import unitcell as UnitCell
from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot3d.SceneWindow import SceneWindow
from silx.io.utils import DataUrl

from ewoks3dxrd.io import read_lattice_cell_data
from ewoks3dxrd.nexus.peaks import read_peaks_attributes
from ewoks3dxrd.nexus.utils import group_exists
from ewoks3dxrd.tasks.grid_index_grains import GridIndexGrains

from .common.ewoks3dxrd_widget import Ewoks3DXRDWidget
from .common.grain_size_slider import GrainSizeSlider
from .common.peak_filter_plot2d import PeakFilterPlot2D
from .common.sphere import Spheres, build_grain_spheres
from .indexer.grid_index_settings import GridIndexSettings


class OWGridIndexing(Ewoks3DXRDWidget, ewokstaskclass=GridIndexGrains):
    name = "Grid Indexing"
    description = "Multiprocessing positional search for the grains."
    icon = "icons/grid-search.svg"
    _ewoks_inputs_to_hide_from_orange = (
        "overwrite",
        "grid_index_parameters",
        "grid_abs_x_limit",
        "grid_abs_y_limit",
        "grid_abs_z_limit",
        "grid_step",
        "seed",
        "analyse_folder",
    )

    class Inputs:
        lattice_file = Input("lattice_file", str)

    def __init__(self):
        super().__init__()
        self._settingsPanelWidget = qt.QWidget(self)
        settingLayout = qt.QVBoxLayout(self._settingsPanelWidget)

        self._settingsPanel = GridIndexSettings(self)
        scrollArea = qt.QScrollArea(self)
        scrollArea.setWidgetResizable(True)
        scrollArea.setWidget(self._settingsPanel)
        settingLayout.addWidget(scrollArea)

        executeOverWriteLayout = qt.QFormLayout()
        self._inputDataUrl = qt.QLineEdit()
        self._overwrite = qt.QCheckBox()
        self._gridIndexerBtn = qt.QPushButton("Run Grid Indexing")
        executeOverWriteLayout.addRow("Input: Data URL", self._inputDataUrl)
        executeOverWriteLayout.addRow("Overwrite", self._overwrite)
        executeOverWriteLayout.addRow(self._gridIndexerBtn)

        self._grainSlider = GrainSizeSlider()
        self._grainSlider.floatValueChanged.connect(self._updateGrainSize)
        self._grainSphere: Optional[Spheres] = None
        executeOverWriteLayout.addRow("Grain Size", self._grainSlider)

        settingLayout.addLayout(executeOverWriteLayout)
        self._gridIndexerBtn.clicked.connect(self.execute_ewoks_task)

        self._grainsPlot = SceneWindow(self)
        self._grainsPlot.setSizePolicy(
            qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding
        )
        self._grainsPlot.getParamTreeView().parent().setVisible(False)
        self._grainsPlot.getGroupResetWidget().parent().setVisible(False)
        self._grainsPlot.getSceneWidget().setBackgroundColor((0.8, 0.8, 0.8, 1.0))

        self._peaksPlot = PeakFilterPlot2D(self)
        self._plotTabWidget = qt.QTabWidget()
        self._plotTabWidget.addTab(self._peaksPlot, "Incoming Peaks")
        self._plotTabWidget.addTab(self._grainsPlot, "Generated Grains")

        self.addControlWidget(self._settingsPanelWidget)
        self.addMainWidget(self._plotTabWidget)

        self._latticeFile = None
        self._ds_ring_limits = 1.4

        self.registerInput(
            "indexer_filtered_data_url", self._getInputUrl, self._inputDataUrl.setText
        )
        self.registerInput(
            "grid_index_parameters",
            self._settingsPanel.getGridIndexParameters,
            self._settingsPanel.setGridIndexParameters,
        )
        self.registerInput(
            "grid_abs_x_limit",
            self._settingsPanel.getGridXLimit,
            self._settingsPanel.setGridXLimit,
        )
        self.registerInput(
            "grid_abs_y_limit",
            self._settingsPanel.getGridYLimit,
            self._settingsPanel.setGridYLimit,
        )
        self.registerInput(
            "grid_abs_z_limit",
            self._settingsPanel.getGridZLimit,
            self._settingsPanel.setGridZLimit,
        )
        self.registerInput(
            "grid_step",
            self._settingsPanel.getGridStep,
            self._settingsPanel.setGridStep,
        )
        self.registerInput(
            "overwrite", self._overwrite.isChecked, self._overwrite.setChecked
        )
        self._restoreDefaultInputs()

    def handleNewSignals(self):
        input_peaks_url = self.get_task_input_value("indexer_filtered_data_url")
        if isinstance(input_peaks_url, str):
            self._inputDataUrl.setText(input_peaks_url)
            self._plotIncomingPeaks()

    def _plotIncomingPeaks(self):
        input_peaks_url = self.get_task_input_value("indexer_filtered_data_url", None)
        if input_peaks_url is None:
            return

        data_url = DataUrl(input_peaks_url)
        nexus_file_path, data_group_path = (
            data_url.file_path(),
            data_url.data_path(),
        )
        incoming_3d_peaks = read_peaks_attributes(
            filename=nexus_file_path,
            process_group=data_group_path,
        )
        self._peaksPlot.addUnfilteredScatter(
            x=incoming_3d_peaks["ds"],
            y=incoming_3d_peaks["eta"],
            value=len(incoming_3d_peaks["eta"]) * [1],
            colormap=Colormap(name="viridis", autoscaleMode="percentile_1_99"),
        )
        self._ds_ring_limits = max(incoming_3d_peaks["ds"])

        self._peaksPlot.setGraphXLabel("Reciprocal distance (ds)")
        self._peaksPlot.setGraphYLabel("Sum Intensity")
        self._peaksPlot.resetZoom()
        if self._latticeFile:
            self._plotLatticeCurve()

    def _plotLatticeCurve(self):
        if self._inputDataUrl.text() == "":
            return
        data_url = DataUrl(self._inputDataUrl.text())
        nexus_file_path, data_group_path = (
            data_url.file_path(),
            data_url.data_path(),
        )
        incoming_3d_peaks = read_peaks_attributes(
            filename=nexus_file_path,
            process_group=data_group_path,
        )

        unit_cell = read_lattice_cell_data(lattice_data_file_path=self._latticeFile)
        unit_cell = UnitCell(
            lattice_parameters=unit_cell.lattice_parameters,
            symmetry=int(unit_cell.space_group),
        )
        unit_cell.makerings(limit=self._ds_ring_limits)
        for i, ring_ds in enumerate(unit_cell.ringds):
            self._peaksPlot.addCurve(
                x=np.array([ring_ds, ring_ds]),
                y=np.array(
                    [
                        min(incoming_3d_peaks["eta"]),
                        max(incoming_3d_peaks["eta"]),
                    ]
                ),
                legend=f"Ring {i}",
                linewidth=2,
            )

    def handleSuccessfulExecution(self):
        output_url = self.get_task_output_value("grid_indexed_grain_data_url", None)
        if output_url is None:
            return
        self._plotTabWidget.setCurrentWidget(self._grainsPlot)
        self._plot3DGrains(output_url)

    def _plot3DGrains(self, url: str):
        self._grainsPlot.getSceneWidget().clearItems()
        self._grainSphere = build_grain_spheres(url, self._grainSlider.value())
        self._grainsPlot.getSceneWidget().addItem(self._grainSphere)

    @Inputs.lattice_file
    def setLatticeFile(self, lattice_file: str):
        self._latticeFile = lattice_file
        self._plotLatticeCurve()

    def _updateGrainSize(self, size: float):
        if self._grainSphere:
            self._grainSphere.setRadiiNorm(radiiNorm=size)
            self._grainSphere.setRadiiNorm(radiiNorm=size)

    def _getInputUrl(self) -> str:
        input_url = DataUrl(self._inputDataUrl.text().strip())
        group_exists(
            filename=input_url.file_path(),
            data_group_path=input_url.data_path(),
        )
        return input_url.path()



from __future__ import annotations
import os
import shutil
import tempfile
from pathlib import Path
from contextlib import contextmanager
from typing import Tuple, Generator
from silx.io.utils import DataUrl

from .nexus.parameters import get_parameters
from .nexus.grains import save_indexed_grains_as_ascii
from .io import (
    load_par_or_cif_file,
    save_geometry_and_lattice_file,
    read_lattice_cell_data,
)


@contextmanager
def tmp_processing_files(
    initial_ubi_file: str | Path, geo_par_file: str | Path, lattice_file: str | Path
) -> Generator[Tuple[str, str], None, None]:
    _, tmp_ubi_file = tempfile.mkstemp()
    _, tmp_par_file = tempfile.mkstemp()

    try:
        shutil.copy2(initial_ubi_file, tmp_ubi_file)

        with open(tmp_par_file, "w") as file:
            geometry_parameters = load_par_or_cif_file(geo_par_file)
            for key, value in geometry_parameters.items():
                file.write(f"{key} {value}\n")

            lattice_parameters = load_par_or_cif_file(lattice_file)
            for key, value in lattice_parameters.items():
                file.write(f"{key} {value}\n")

        yield tmp_ubi_file, tmp_par_file
    finally:
        os.remove(tmp_ubi_file)
        os.remove(tmp_par_file)


@contextmanager
def tmp_grain_processing_files(
    ubi_init_data_url: str, geo_init_data_url: str, lattice_parameter_file: str
) -> Generator[Tuple[str, str], None, None]:
    _, tmp_ubi_file = tempfile.mkstemp()
    _, tmp_par_file = tempfile.mkstemp()

    try:
        ubi_url = DataUrl(ubi_init_data_url)
        save_indexed_grains_as_ascii(
            grain_file_h5=ubi_url.file_path(),
            process_group_path=ubi_url.data_path(),
            grain_file_ascii=tmp_ubi_file,
        )

        geo_url = DataUrl(geo_init_data_url)
        geo_par_dict = get_parameters(
            filename=geo_url.file_path(),
            process_group_name=geo_url.data_path(),
        )
        unit_cell_parameters = read_lattice_cell_data(
            lattice_data_file_path=lattice_parameter_file
        )
        save_geometry_and_lattice_file(
            file_path=tmp_par_file,
            geom_dict=geo_par_dict,
            lattice_dict={
                "lattice_parameters": unit_cell_parameters.lattice_parameters,
                "lattice_space_group": unit_cell_parameters.space_group,
            },
        )

        yield tmp_ubi_file, tmp_par_file
    finally:
        os.remove(tmp_ubi_file)
        os.remove(tmp_par_file)


@contextmanager
def tmp_lattice_and_geo_file(
    lat_par_data_url: str, geo_par_data_url: str
) -> Generator[Path, None, None]:
    with tempfile.TemporaryDirectory() as tmp_dir:
        geo_url = DataUrl(geo_par_data_url)
        geo_par_dict = get_parameters(
            filename=geo_url.file_path(),
            process_group_name=geo_url.data_path(),
        )
        lat_url = DataUrl(lat_par_data_url)
        lat_par_dict = get_parameters(
            filename=lat_url.file_path(),
            process_group_name=lat_url.data_path(),
        )
        tmp_par_file = Path(tmp_dir) / "tmp.par"
        save_geometry_and_lattice_file(
            file_path=tmp_par_file,
            geom_dict=geo_par_dict,
            lattice_dict=lat_par_dict,
        )
        yield tmp_par_file



from pathlib import Path

from ewokscore import Task
from pydantic import Field

from ..io import copy_par_file
from ..models import InputsWithOverwrite
from ..nexus.peaks import read_peaks_attributes, save_nexus_process
from ..nexus.utils import (
    check_throw_write_data_group_url,
    get_data_url_paths,
    get_entry_name,
)
from ..utils import update_geometry


class Inputs(InputsWithOverwrite):
    spatial_corrected_data_url: str = Field(
        description="Path to the NXprocess group containing the detector spatial corrected 3D peaks"
    )
    geometry_par_file: str = Field(
        description="Path to the `.par` file containing geometry parameters"
    )


class GeometryTransformation(
    Task,
    input_model=Inputs,
    output_names=["geometry_updated_data_url", "wavelength"],
):
    """
    This task performs the following operations:
    1. Gathers geometry information from the `geometry_tdxrd.par` file.
    2. Copy the geometry file in the directory structure:
            `analysis_folder / dset_name / (dset_name + sample_name)`.
            i.e the parent folder of 'detector_spatial_corrected_3d_peaks_file'
    3. Applies the geometry correction to the 3D peaks column file using
        the `ImageD11` dataset class.
    4. Save the geometry corrected 3d peaks columnfile (output: `geometry_updated_3d_peaks_file`)
    """

    def run(self):
        inputs = Inputs(**self.get_input_values())
        nexus_file_path, spatial_corrected_data_group_path = get_data_url_paths(
            inputs.spatial_corrected_data_url
        )

        entry_name = get_entry_name(spatial_corrected_data_group_path)
        geo_updated_data_group = f"{entry_name}/geometry_updated_peaks"
        check_throw_write_data_group_url(
            inputs.overwrite, nexus_file_path, geo_updated_data_group
        )

        sp_corrected_3d_peaks = read_peaks_attributes(
            filename=nexus_file_path,
            process_group=spatial_corrected_data_group_path,
        )
        geometry_file = copy_par_file(
            inputs.geometry_par_file, Path(nexus_file_path).parent
        )
        peaks_data, geom_parameters = update_geometry(
            sp_corrected_3d_peaks, geometry_file
        )

        nxprocess_url = save_nexus_process(
            filename=nexus_file_path,
            entry_name=entry_name,
            process_name="geometry_updated_peaks",
            peaks_data=peaks_data,
            config_settings=geom_parameters,
            pks_axes=("ds", "eta"),
            overwrite=inputs.overwrite,
            ln_pks_from_group_name=spatial_corrected_data_group_path,
        )
        self.outputs.geometry_updated_data_url = nxprocess_url
        self.outputs.wavelength = geom_parameters["wavelength"]



from __future__ import annotations

from pathlib import Path
from typing import Any, List, Tuple

import h5py
import numpy as np
from ImageD11 import grain as grainmod
from ImageD11.grain import grain as Grain
from silx.io.utils import h5py_read_dataset

from .parameters import create_parameters_group


def create_nexus_grains_data_group(
    group_path: h5py.Group,
    grains: List[Grain],
) -> str:

    if not grains:
        raise ValueError("Grains list is empty!")

    grain_data_group = group_path.require_group("grains")
    grain_data_group.attrs["NX_class"] = "NXdata"
    num_grains = len(grains)

    ubi_matrices = grain_data_group.create_dataset(
        "UBI", shape=(num_grains, 3, 3), dtype=grains[0].ubi.dtype
    )
    translations = grain_data_group.create_dataset(
        "translation", shape=(num_grains, 3), dtype=grains[0].ubi.dtype
    )
    npks = grain_data_group.create_dataset("npks", shape=(num_grains,), dtype=np.int64)
    nuniq = grain_data_group.create_dataset(
        "nuniq", shape=(num_grains,), dtype=np.int64
    )
    names = grain_data_group.create_dataset(
        "name", shape=(num_grains,), dtype=h5py.string_dtype(encoding="utf-8")
    )
    intensity_infos = grain_data_group.create_dataset(
        "intensity_info", shape=(num_grains,), dtype=h5py.string_dtype(encoding="utf-8")
    )
    mean_intensity = grain_data_group.create_dataset(
        "mean_intensity", shape=(num_grains,), dtype=np.float64
    )

    for i, grain in enumerate(grains):
        ubi_matrices[i] = grain.ubi
        if hasattr(grain, "translation"):
            translations[i] = grain.translation
        if hasattr(grain, "npks"):
            npks[i] = grain.npks
        if hasattr(grain, "nuniq"):
            nuniq[i] = grain.nuniq
        if hasattr(grain, "name"):
            names[i] = grain.name
        if hasattr(grain, "intensity_info"):
            intensity_infos[i] = grain.intensity_info
            if "no peaks" in grain.intensity_info:
                mean_intensity[i] = 0
            else:
                mean_intensity[i] = float(
                    grain.intensity_info.split("mean = ")[1]
                    .split(" , ")[0]
                    .replace("'", "")
                )

    return grain_data_group.name


def save_nexus_grain_process(
    filename: str | Path,
    entry_name: str,
    process_name: str,
    config_settings: dict[str, Any] | dict[str, dict[str, Any]],
    grains: List[Grain],
    overwrite: bool = False,
) -> str:
    with h5py.File(filename, "a") as h5_file:
        entry = h5_file.require_group(entry_name)
        entry.attrs["NX_class"] = "NXentry"
        if process_name in entry:
            if overwrite:
                del entry[process_name]
            else:
                raise FileExistsError(
                    f"""In the nexus process file name {filename}, there is already a
                    nexus process group {process_name} exists.
                    To overwrite provide a overwrite permission.
                    """
                )

        process_group = entry.create_group(process_name)
        process_group.attrs["NX_class"] = "NXprocess"
        create_nexus_grains_data_group(
            group_path=process_group,
            grains=grains,
        )
        create_parameters_group(
            group_path=process_group,
            config_settings=config_settings,
        )
        return f"{filename}::{process_group.name}"


def create_nexus_ubi(
    grains: List[Grain],
    entry_name: str,
    grain_file: str | Path,
    grain_settings: dict[str, Any],
    peaks_path: Tuple[str | Path, str],
):
    """
    Create nexus file with entry_name place the grains (list of UBIs)
    and also link the peaks used to generate these UBIS as external link
    """
    with h5py.File(grain_file, "w") as gf:
        entry = gf.create_group(entry_name)
        entry.attrs["NX_class"] = "NXentry"
        grain_group = entry.create_group("indexed_grains")
        grain_group.attrs["NX_class"] = "NXprocess"
        grain_group["peaks"] = h5py.ExternalLink(peaks_path[0], peaks_path[1])
        ubi_group = grain_group.create_group("UBI")
        ubi_group.attrs["NX_class"] = "NXdata"
        ubi_matrices = ubi_group.create_dataset(
            "UBI", shape=(len(grains), 3, 3), dtype=grains[0].ubi.dtype
        )
        for i, grain in enumerate(grains):
            ubi_matrices[i] = grain.ubi
        create_parameters_group(
            parent_group=grain_group, config_settings=grain_settings
        )


def create_nexus_grains(
    grains: List[Grain],
    entry_name: str,
    grain_file: str | Path,
    grain_settings: dict[str, Any],
    grain_group_name: str,
):
    num_grains = len(grains)
    with h5py.File(grain_file, "a") as gf:
        entry = gf[entry_name]
        grp = entry.create_group(grain_group_name)
        grp.attrs["NX_class"] = "NXprocess"
        grains_gr = grp.create_group("grains")
        grains_gr.attrs["NX_class"] = "NXdata"

        ubi_matrices = grains_gr.create_dataset(
            "UBI", shape=(num_grains, 3, 3), dtype=grains[0].ubi.dtype
        )
        translations = grains_gr.create_dataset(
            "translation", shape=(num_grains, 3), dtype=grains[0].ubi.dtype
        )
        npks = grains_gr.create_dataset("npks", shape=(num_grains,), dtype=np.int64)
        nuniq = grains_gr.create_dataset("nuniq", shape=(num_grains,), dtype=np.int64)
        for i, grain in enumerate(grains):
            ubi_matrices[i] = grain.ubi
            translations[i] = grain.translation
            npks[i] = grain.npks
            nuniq[i] = grain.nuniq

        create_parameters_group(parent_group=grp, config_settings=grain_settings)


def read_grains(
    grain_file_h5: str | Path, entry_name: str, process_group_name: str
) -> List[Grain]:

    grains_list = []
    with h5py.File(grain_file_h5, "r") as gf:
        entry = gf[f"{entry_name}/{process_group_name}/grains"]
        for i in range(entry["translation"][()].shape[0]):
            gr = Grain(ubi=entry["UBI"][i], translation=entry["translation"][i])
            gr.npks = entry["npks"][i]
            gr.nuniq = entry["nuniq"][i]
            gr.name = entry["name"][i]
            gr.intensity_info = h5py_read_dataset(entry["intensity_info"], index=i)
            gr.mean_intensity = entry["mean_intensity"][i]
            grains_list.append(gr)
    return grains_list


def read_grains_attributes(filename: str | Path, process_group: str) -> dict[str, list]:
    with h5py.File(filename, "r") as h5file:
        grain_group = h5file[f"{process_group}/grains"]
        return {name: dataset[()] for name, dataset in grain_group.items()}


def save_indexed_grains_as_ascii(
    grain_file_h5: str | Path,
    process_group_path: str,
    grain_file_ascii: str | Path,
):
    """
    Function to extract grains that generated by Indexing function,
    which is simpler, i.e it only has UBI matrices.
    """
    grains_list = []
    with h5py.File(grain_file_h5, "r") as gf:
        ubi_dataset = gf[f"{process_group_path}/grains/UBI"]
        for ubi_matrix in ubi_dataset:
            grains_list.append(Grain(ubi=ubi_matrix))

    grainmod.write_grain_file(grain_file_ascii, grains_list)



from __future__ import annotations

from typing import Tuple

from silx.gui import qt

TOLERANCE_PRESETS: dict[str, Tuple[float, ...]] = {
    "Fast": (0.02,),
    "Moderate": (0.02, 0.015, 0.01),
    "Fine": (0.02, 0.015, 0.0125, 0.01, 0.007),
}


class ToleranceSelector(qt.QWidget):

    def __init__(self, parent=None):
        super().__init__(parent)

        layout = qt.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        self._radioFast = qt.QRadioButton("Fast")
        self._radioModerate = qt.QRadioButton("Moderate")
        self._radioFine = qt.QRadioButton("Fine")
        self._radioModerate.setChecked(True)

        layout.addWidget(self._radioFast)
        layout.addWidget(self._radioModerate)
        layout.addWidget(self._radioFine)

    def getValue(self) -> Tuple[float, ...]:
        if self._radioFast.isChecked():
            return TOLERANCE_PRESETS["Fast"]
        elif self._radioModerate.isChecked():
            return TOLERANCE_PRESETS["Moderate"]
        elif self._radioFine.isChecked():
            return TOLERANCE_PRESETS["Fine"]
        else:
            raise ValueError("No tolerance preset selected")

    def setValue(self, value: Tuple[float, ...]):
        if value == TOLERANCE_PRESETS["Fine"]:
            self._radioFine.setChecked(True)
        if value == TOLERANCE_PRESETS["Moderate"]:
            self._radioModerate.setChecked(True)
        else:
            self._radioFast.setChecked(True)
            self._radioFast.setChecked(True)
            self._radioFast.setChecked(True)
            self._radioFast.setChecked(True)



from typing import Optional
from silx.gui import qt


class QDoubleSlider(qt.QSlider):
    floatValueChanged = qt.Signal(float)

    def __init__(
        self, orientation: qt.Qt.Orientation, parent: Optional[qt.QWidget] = None
    ):
        super().__init__(orientation, parent)
        self._min = 0.0
        self._max = 1.0
        self._precision = 10000
        super().setRange(0, self._precision)
        self.valueChanged.connect(self._emitFloatValue)

    def setRange(self, min_val: float, max_val: float):
        self._min = min_val
        self._max = max_val

    def value(self) -> float:
        int_val = super().value()
        return self._min + (self._max - self._min) * int_val / self._precision

    def setFloatValue(self, val: float):
        ratio = (val - self._min) / (self._max - self._min)
        int_val = int(ratio * self._precision)
        super().setValue(int_val)

    def _emitFloatValue(self, _):
        float_val = self.value()
        self.floatValueChanged.emit(float_val)



from ewoks import Task
from pydantic import Field

from ..filtering import filter_by_intensity
from ..models import InputsWithOverwrite
from ..nexus.peaks import read_peaks_attributes, save_nexus_process
from ..nexus.utils import (
    check_throw_write_data_group_url,
    get_data_url_paths,
    get_entry_name,
)


class Inputs(InputsWithOverwrite):
    lattice_filtered_data_url: str = Field(
        description="Path to the NXprocess group containing the peaks corrected for geometry and detector"
    )
    intensity_frac: float = Field(
        ge=0,
        le=1,
        description="Filter peaks whose intensity is below this value",
    )
    process_group_name: str = Field(
        default="intensity_filtered_peaks",
        description="NXprocess group name where to save the filtered peaks",
    )


class FilterByIntensity(
    Task, input_model=Inputs, output_names=["intensity_filtered_data_url"]
):
    """
    Does the Intensity based peaks filter,
        computes intensity metric based on sum_intensity, ds (reciprocal distance) columns from the input file
        normalize with the maximum value of intensity metric,
        only keeps the rows whose value is above the given input 'intensity_frac'
        and save them in .h5 format.

    Outputs:

    - `intensity_filtered_data_url` (str): Data path to the intensity filtered 'NxProcess group data' peaks
    """

    def run(self):
        inputs = Inputs(**self.get_input_values())
        nexus_file_path, lattice_filtered_data_group_path = get_data_url_paths(
            inputs.lattice_filtered_data_url
        )

        entry_name = get_entry_name(lattice_filtered_data_group_path)

        check_throw_write_data_group_url(
            inputs.overwrite,
            nexus_file_path,
            f"{entry_name}/{inputs.process_group_name}",
        )

        filtered_3d_peaks_dict = filter_by_intensity(
            peaks=read_peaks_attributes(
                filename=nexus_file_path, process_group=lattice_filtered_data_group_path
            ),
            intensity_frac=inputs.intensity_frac,
        )

        nxprocess_url = save_nexus_process(
            filename=nexus_file_path,
            entry_name=entry_name,
            process_name=inputs.process_group_name,
            peaks_data=filtered_3d_peaks_dict,
            config_settings={
                "intensity_frac": inputs.intensity_frac,
                "data_from": f"{nexus_file_path}::{lattice_filtered_data_group_path}",
            },
            pks_axes=("ds", "eta"),
            overwrite=inputs.overwrite,
        )
        self.outputs.intensity_filtered_data_url = nxprocess_url



from __future__ import annotations

from typing import Any

from silx.gui import qt

from .lattice_parameters import LatticeParameters


class LatticeFilteringSettings(qt.QWidget):
    sigPlotLatticeRings = qt.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Lattice Config")
        self.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding)
        layout = qt.QVBoxLayout()
        self._latticeParGroup = LatticeParameters(parent=self)
        layout.addWidget(self._latticeParGroup)
        self._latticeParGroup.sigLatticeParamsChanged.connect(
            self.sigPlotLatticeRings.emit
        )

        setupGroup = qt.QGroupBox("Tolerance Settings", parent=self)
        setupLayout = qt.QFormLayout(setupGroup)
        self._reciprocalDistTol = qt.QDoubleSpinBox()
        self._reciprocalDistTol.setRange(0.0001, 1)
        self._reciprocalDistTol.setValue(0.005)
        self._reciprocalDistTol.setDecimals(4)
        setupLayout.addRow("Reciprocal Distance Tolerance:", self._reciprocalDistTol)

        self._reciprocalDistMaxCheckbox = qt.QCheckBox("Max Reciprocal Distance")
        self._reciprocalDistMaxCheckbox.setChecked(False)
        self._reciprocalDistMax = qt.QDoubleSpinBox()
        self._reciprocalDistMax.setRange(0.0001, 2)
        self._reciprocalDistMax.setValue(2)
        self._reciprocalDistMax.setDecimals(4)
        self._reciprocalDistMax.setEnabled(False)
        setupLayout.addRow(self._reciprocalDistMaxCheckbox, self._reciprocalDistMax)

        layout.addWidget(setupGroup)
        self.setLayout(layout)
        self._reciprocalDistMaxCheckbox.toggled.connect(
            self._reciprocalDistMax.setEnabled
        )

    def getLatticeParameters(self) -> dict[str, Any]:
        return self._latticeParGroup.getLatticeParameters()

    def setLatticeFile(self, value: str):
        return self._latticeParGroup.setFile(value)

    def getReciprocalDistTol(self) -> float:
        return self._reciprocalDistTol.value()

    def setReciprocalDistTol(self, value: float):
        self._reciprocalDistTol.setValue(value)

    def getReciprocalDistMax(self) -> float | None:
        return (
            self._reciprocalDistMax.value()
            if self._reciprocalDistMaxCheckbox.isChecked()
            else None
        )

    def setReciprocalDistMax(self, value: float | None):
        if value is None:
            self._reciprocalDistMax.clear()
            self._reciprocalDistMaxCheckbox.setChecked(False)
        else:
            self._reciprocalDistMax.setValue(value)
            self._reciprocalDistMaxCheckbox.setChecked(True)



import pytest

from ewoks3dxrd.io import read_lattice_cell_data
from ewoks3dxrd.io import get_monitor_scale_factor


@pytest.mark.parametrize("space_group", (229, "I"))
def test_read_lattice_cell_data(tmp_path, space_group):
    par_file = tmp_path / "Fe.par"
    with open(par_file, "w") as f:
        f.write(
            f"""cell__a 2.8694
cell__b 2.8694
cell__c 2.8694
cell_alpha 90.0
cell_beta 90.0
cell_gamma 90.0
cell_lattice_[P,A,B,C,I,F,R] {space_group}
"""
        )

    unit_cell_parameters = read_lattice_cell_data(par_file)

    assert unit_cell_parameters.a == 2.8694
    assert unit_cell_parameters.b == 2.8694
    assert unit_cell_parameters.c == 2.8694
    assert unit_cell_parameters.alpha == 90
    assert unit_cell_parameters.beta == 90
    assert unit_cell_parameters.gamma == 90
    assert unit_cell_parameters.space_group == space_group


def test_get_monitor_scale_factor_id03():
    scale_factor = get_monitor_scale_factor(
        masterfile_path="/data/projects/id03_3dxrd/id03_expt/RAW_DATA/Al_1050/Al_1050_rot_4/Al_1050_rot_4.h5",
        scan_number=1,
        monitor_name="pico4",
    )
    assert scale_factor.shape == (1800,)

    with pytest.raises(KeyError):
        scale_factor = get_monitor_scale_factor(
            masterfile_path="/data/projects/id03_3dxrd/id03_expt/RAW_DATA/Al_1050/Al_1050_rot_4/Al_1050_rot_4.h5",
            scan_number=1,
            monitor_name="not_existing",
        )


def test_get_monitor_scale_factor_id11():
    scale_factor = get_monitor_scale_factor(
        masterfile_path="/data/projects/id03_3dxrd/expt/RAW_DATA/FeAu_0p5_tR/FeAu_0p5_tR_ff1/FeAu_0p5_tR_ff1.h5",
        scan_number=1,
        monitor_name="fpico4",
    )
    assert scale_factor.shape == (1440,)



import os
import shutil
from pathlib import Path

import pytest
from ewoksorange.canvas.handler import OrangeCanvasHandler
from ewoksorange.tests.conftest import qtapp  # noqa F401

from ..nexus.grains import read_grains_attributes
from ..nexus.utils import get_data_url_paths


@pytest.fixture(scope="session")
def ewoks_orange_canvas(qtapp):  # noqa F811
    with OrangeCanvasHandler() as handler:
        yield handler


@pytest.fixture
def inp_config(tmp_path):
    master_file = "/data/projects/id03_3dxrd/expt/RAW_DATA/FeAu_0p5_tR/FeAu_0p5_tR_ff1/FeAu_0p5_tR_ff1.h5"

    if not os.path.exists(master_file):
        raise FileNotFoundError(
            f"""
            Could not find {master_file}.
            Before running this test, be sure to have a link to /data/projects at /data/:
             - mkdir -p /data/projects/id03_3dxrd
             - ln -s /gpfs/.../data/projects /data
            Sudo rights might be needed.
            """
        )

    return {
        "detector": "frelon3",
        "omega_motor": "diffrz",
        "dty_motor": "diffty",
        "bg_file": None,
        "mask_file": "/data/projects/id03_3dxrd/expt/PROCESSED_DATA/mask.edf",
        "spline_file": "/data/projects/id03_3dxrd/expt/PROCESSED_DATA/frelon36.spline",
        "e2dx_file": None,
        "e2dy_file": None,
        "master_file": master_file,
        "scan_number": 1,
        "analyse_folder": os.path.join(tmp_path, "test_my_task"),
        "stateful_imageD11_file": None,
    }


@pytest.fixture
def id03_inp_config(tmp_path):
    master_file = "/data/projects/id03_3dxrd/id03_expt/RAW_DATA/Al_1050/Al_1050_rot_4/Al_1050_rot_4.h5"

    if not os.path.exists(master_file):
        raise FileNotFoundError(
            f"""
            Could not find {master_file}.
            Before running this test, be sure to have a link to /data/projects at /data/:
             - mkdir -p /data/projects/id03_3dxrd
             - ln -s /gpfs/.../data/projects /data
            Sudo rights might be needed.
            """
        )

    return {
        "detector": "frelon1",
        "omega_motor": "omega",
        "dty_motor": "diffty",
        "bg_file": "/data/projects/id03_3dxrd/id03_expt/PROCESSED_DATA/bkg_avg.edf",
        "mask_file": "/data/projects/id03_3dxrd/id03_expt/PROCESSED_DATA/frelon1_mask_20250122.edf",
        "spline_file": "/data/projects/id03_3dxrd/id03_expt/PROCESSED_DATA/distortion_frelon.spline",
        "e2dx_file": None,
        "e2dy_file": None,
        "master_file": master_file,
        "scan_number": 1,
        "analyse_folder": os.path.join(tmp_path, "test_my_task"),
        "stateful_imageD11_file": None,
    }


@pytest.fixture()
def nexus_3dxrd_file(tmp_path):
    new_filename = tmp_path / "nexus_segment.h5"
    shutil.copy2(
        "/data/projects/id03_3dxrd/ewoks_test_data/nexus_segment.h5",
        new_filename,
    )

    return new_filename


def assert_indexing_results(indexing_task_outputs):
    nexus_file_path, ubi_data_group_path = get_data_url_paths(
        indexing_task_outputs["indexed_grain_data_url"]
    )

    assert Path(nexus_file_path).exists()
    dict_grains = read_grains_attributes(
        filename=nexus_file_path, process_group=ubi_data_group_path
    )
    assert dict_grains["UBI"].shape == (59, 3, 3)


def assert_grain_map_results(grain_map_task_outputs):
    nexus_file_path, ubi_data_group_path = get_data_url_paths(
        grain_map_task_outputs["make_map_data_url"]
    )
    assert Path(nexus_file_path).exists()
    dict_grains = read_grains_attributes(
        filename=nexus_file_path, process_group=ubi_data_group_path
    )
    assert dict_grains["UBI"].shape == (52, 3, 3)



from ewoks3dxrd.tasks.index_grains import IndexGrains

from .conftest import assert_indexing_results


def test_indexing(nexus_3dxrd_file):
    filepath = f"{nexus_3dxrd_file}::/1.1/intensity_filtered_inner_peaks"

    inputs = {
        "intensity_filtered_data_url": filepath,
        "reciprocal_dist_tol": 0.05,
        "rings": (0, 1),
        "scoring_rings": (0, 1, 2, 3),
        "hkl_tols": (0.01, 0.02, 0.03, 0.04),
        "min_pks_frac": (0.9, 0.75),
        "overwrite": True,
        "wavelength": 0.28457,
        "lattice_file": "/data/projects/id03_3dxrd/expt/PROCESSED_DATA/Fe.par",
    }

    task = IndexGrains(inputs=inputs)
    task.execute()

    assert_indexing_results(task.outputs)



import sys
from pathlib import Path
from typing import Literal, Optional, Sequence, Tuple, Union

from ewoks import BaseInputModel

if sys.version_info < (3, 9):
    from typing_extensions import Annotated
else:
    from typing import Annotated

from pydantic import AliasChoices, BaseModel, Field, field_validator

DetectorCorrectionFiles = Union[Tuple[str, str], str]

Translations = Tuple[Tuple[int, int, int], ...]


class SampleConfig(BaseModel):
    omega_motor: Literal["diffrz", "omega"]
    master_file: str
    scan_number: int

    @field_validator("master_file")
    @classmethod
    def path_exists(cls, path: Optional[str]):
        if path is None or not Path(path).exists():
            raise ValueError(f"Supplied path {path} does not exist.")
        return path


class SegmenterFolderConfig(SampleConfig):
    detector: Literal["frelon1", "frelon3"]
    analyse_folder: str


class SegmenterConfig(BaseModel):
    threshold: int
    smooth_sigma: float
    bgc: float
    min_px: int
    offset_threshold: int
    ratio_threshold: int


class SegmenterCorrectionFiles(BaseModel):
    bg_file: Optional[str] = None
    mask_file: Optional[str] = None
    dark_file: Optional[str] = None
    flat_file: Optional[str] = None

    @field_validator("bg_file", "mask_file", "dark_file", "flat_file")
    @classmethod
    def path_exists(cls, path: Optional[str]):
        if path is not None and not Path(path).exists():
            raise ValueError(f"Supplied path {path} does not exist.")
        return path


class UnitCellParameters(BaseModel):
    a: float = Field(validation_alias=AliasChoices("cell__a", "_cell_length_a"))
    b: float = Field(validation_alias=AliasChoices("cell__b", "_cell_length_b"))
    c: float = Field(validation_alias=AliasChoices("cell__b", "_cell_length_b"))
    alpha: float = Field(
        validation_alias=AliasChoices("cell_alpha", "_cell_angle_alpha")
    )
    beta: float = Field(validation_alias=AliasChoices("cell_beta", "_cell_angle_beta"))
    gamma: float = Field(
        validation_alias=AliasChoices("cell_gamma", "_cell_angle_gamma")
    )
    space_group: Union[
        Literal["P", "A", "B", "C", "I", "F", "R"], Annotated[int, Field(ge=1, le=230)]
    ] = Field(
        validation_alias=AliasChoices(
            "cell_lattice_[P,A,B,C,I,F,R]", "_space_group_IT_number"
        )
    )

    @property
    def lattice_parameters(self) -> Tuple[float, float, float, float, float, float]:
        return self.a, self.b, self.c, self.alpha, self.beta, self.gamma


class GridIndexParameters(BaseModel):
    """
    The variable name defined here are following the same as used by the grid_index_parallel module from ImageD11 library.
    https://github.com/FABLE-3DXRD/ImageD11/blob/master/ImageD11/grid_index_parallel.py
    """

    NPKS: int
    DSTOL: float = 0.004
    RING1: Sequence[int] = (1, 0)
    RING2: Sequence[int] = (0,)
    NUL: bool = True
    FITPOS: bool = True
    tolangle: float = 0.50
    toldist: float = 100.0
    # by providing None, here, ImageD11 will use all the available cores from the server/system
    NPROC: Optional[int] = None
    NTHREAD: int = 1
    COSTOL: Optional[float] = None
    OMEGAFLOAT: Optional[float] = None
    TOLSEQ: Tuple[float, ...] = (0.02, 0.015, 0.01)
    SYMMETRY: str = "cubic"


class InputsWithOverwrite(BaseInputModel):
    overwrite: bool = Field(
        default=False, description="Overwrite NXprocess group if it already exists"
    )



from __future__ import annotations

from pathlib import Path
from typing import Tuple, Optional
import numpy as np
import random
from ImageD11 import blobcorrector
from ImageD11 import columnfile as PeakColumnFile
from ImageD11.columnfile import columnfile as ColumnFile
from ImageD11.refinegrains import refinegrains as RefineGrains

from .models import DetectorCorrectionFiles


def do_spatial_correction(
    peak_3d_dict: dict, correction_files: DetectorCorrectionFiles
) -> ColumnFile:
    peak_3d_dict["spot3d_id"] = np.arange(len(peak_3d_dict["s_raw"]))
    raw_columnfile_3d = PeakColumnFile.colfile_from_dict(peak_3d_dict)

    if isinstance(correction_files, str):
        spline_file = correction_files
        return blobcorrector.correct_cf_with_spline(raw_columnfile_3d, spline_file)

    if isinstance(correction_files, tuple) and len(correction_files) == 2:
        e2dx_file, e2dy_file = correction_files
        return blobcorrector.correct_cf_with_dxdyfiles(
            raw_columnfile_3d, e2dx_file, e2dy_file
        )

    raise ValueError(
        f"Detector Spatial correction cannot be performed unless a spline file or a couple of e2dx, ed2dy files is provided. Got {correction_files}."
    )


def refine_grains(
    tolerance: float,
    intensity_tth_range: Tuple[float, float],
    omega_slop: float,
    symmetry: str,
    parameter_file: str | Path,
    filtered_peaks_file: str | Path,
    ubi_file: str | Path,
) -> RefineGrains:
    refined_grains = RefineGrains(
        tolerance=tolerance,
        intensity_tth_range=intensity_tth_range,
        OmSlop=omega_slop,
    )
    refined_grains.loadparameters(str(parameter_file))
    refined_grains.loadfiltered(str(filtered_peaks_file))
    refined_grains.readubis(str(ubi_file))
    refined_grains.makeuniq(symmetry)
    refined_grains.generate_grains()
    refined_grains.refinepositions()

    return refined_grains


def generate_translations(
    max_x: int, max_y: int, max_z: int, step: int, seed: Optional[int] = None
) -> Tuple[Tuple[int, int, int], ...]:
    translations = [
        (t_x, t_y, t_z)
        for t_x in range(-max_x, max_x + 1, step)
        for t_y in range(-max_y, max_y + 1, step)
        for t_z in range(-max_z, max_z + 1, step)
    ]

    if seed is None:
        return tuple(translations)
    else:
        rnd = random.Random(seed)
        rnd.shuffle(translations)
        return tuple(translations)


def update_geometry(peaks: dict, geometry_file: Path) -> Tuple[dict, dict]:
    cf = PeakColumnFile.colfile_from_dict(peaks)
    # The following code derives from `update_colfile_pars`:
    # https://github.com/FABLE-3DXRD/ImageD11/blob/a27e67ffbd8471a7eb036ce7661d006027739cb6/ImageD11/sinograms/dataset.py#L870
    cf.parameters.loadparameters(filename=str(geometry_file))
    # updateGeometry() adds various geometry parameters from the geometry .par file into the column file
    cf.updateGeometry()

    return {
        key: cf[key] for key in ("ds", "eta", "gx", "gy", "gz", "tth", "xl", "yl", "zl")
    }, cf.parameters.parameters



from __future__ import annotations

import numpy as np
import qtawesome
from silx.gui import qt
from silx.gui.colors import Colormap, rgba
from silx.gui.dialog.ColormapDialog import ColormapDialog
from silx.gui.plot import PlotWidget, actions
from silx.gui.plot.ColorBar import ColorBarWidget
from silx.gui.plot.items import Scatter
from silx.gui.plot.tools import PositionInfo
from silx.gui.plot.utils.axis import SyncAxes

from .peak_color_combobox import PeakColorComboBox

IMAGE_NAME = "image"
SCATTER_NAME = "scatter"


class _PlotPanel(qt.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

        self._plot = PlotWidget(parent=self, backend="gl")
        self._plot.setGraphGrid(True)
        self._plot.setAxesMargins(0.05, 0.03, 0.03, 0.05)
        self._plot.setKeepDataAspectRatio(True)
        cmap = Colormap(
            name="viridis", normalization="log", autoscaleMode="percentile_1_99"
        )
        self._plot.setDefaultColormap(cmap)

        self._colorbar = ColorBarWidget(parent=self)
        self._colorbar.setPlot(self._plot)
        self._colorbar.setSizePolicy(qt.QSizePolicy.Fixed, qt.QSizePolicy.Expanding)
        self._colorbar.setVisible(False)

        positionInfo = PositionInfo(
            plot=self._plot,
            parent=self,
            converters=(
                ("X", lambda x, y: x),
                ("Y", lambda x, y: y),
                ("Data", self._dataConverter),
            ),
        )

        layout = qt.QGridLayout(self)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._plot, 0, 0)
        layout.addWidget(self._colorbar, 0, 1)
        layout.addWidget(positionInfo, 1, 0, 1, 2)
        self.setLayout(layout)

    def _dataConverter(self, x: float, y: float) -> str:
        # Inspired by https://github.com/silx-kit/silx/blob/a9d04d831bd6dcb88026aac896cf834a41d0fc93/src/silx/gui/plot/PlotWindow.py#L999
        image = self._plot.getImage(IMAGE_NAME)
        if image is None:
            return "-"

        pixel_coords = self._plot.dataToPixel(x, y)
        if pixel_coords is None:
            return "-"

        picked_image = image.pick(*pixel_coords)
        if picked_image is None:
            return "-"

        indices = picked_image.getIndices(copy=False)
        if indices is None:
            return "-"

        row, col = indices[0][0], indices[1][0]
        return f"{image.getData(copy=False)[row, col]:.2f}"

    def getPlotWidget(self) -> PlotWidget:
        return self._plot

    def setColorBarVisible(self, visible: bool) -> None:
        self._colorbar.setVisible(visible)

    def setImage(
        self, array2D: np.ndarray, title: str, xLabel: str, yLabel: str
    ) -> None:
        self._plot.setGraphTitle(title)
        # Reset zoom only if the image does not exist
        resetzoom = self._plot.getImage(IMAGE_NAME) is None
        self._plot.addImage(array2D, legend=IMAGE_NAME, resetzoom=resetzoom)
        self._plot.setGraphXLabel(xLabel)
        self._plot.setGraphYLabel(yLabel)

    def setScatter(
        self, x: np.ndarray, y: np.ndarray, colormap: Colormap | dict, symbol: str
    ):
        self._plot.addScatter(
            x=x,
            y=y,
            value=np.ones(len(x)),
            colormap=colormap,
            symbol=symbol,
            legend=SCATTER_NAME,
        )

    def getScatter(self) -> Scatter | None:
        return self._plot.getScatter(legend=SCATTER_NAME)

    def getColorBarWidget(self) -> ColorBarWidget:
        return self._colorbar

    def setScatterColor(self, color: str):
        scatter = self._plot.getScatter(SCATTER_NAME)
        if scatter is None:
            return
        scatter.setColormap({"colors": [rgba(color)]})


class SilxSyncDualPlot(qt.QMainWindow):

    def __init__(self, parent=None):
        super().__init__(parent)
        self._initCentralLayout()
        self._addToolbars()
        self._syncAxes()

    def _initCentralLayout(self):
        self._leftPanel = _PlotPanel(self)
        self._rightPanel = _PlotPanel(self)

        splitter = qt.QSplitter(qt.Qt.Horizontal)
        splitter.addWidget(self._leftPanel)
        splitter.addWidget(self._rightPanel)
        self.setCentralWidget(splitter)

    def _addToolbars(self):
        toolbar = qt.QToolBar("Plot Tools", self)
        toolbar.setToolButtonStyle(qt.Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        self.addToolBar(qt.Qt.TopToolBarArea, toolbar)

        zoomAction = actions.mode.ZoomModeAction(
            parent=self, plot=self._leftPanel.getPlotWidget()
        )
        panAction = actions.mode.PanModeAction(
            parent=self, plot=self._leftPanel.getPlotWidget()
        )
        toolbar.addAction(panAction)
        toolbar.addAction(zoomAction)
        self._leftPanel.getPlotWidget().setInteractiveMode("pan")
        self._leftPanel.getPlotWidget().sigInteractiveModeChanged.connect(
            self._syncInteractionMode
        )

        resetZoom = actions.control.ResetZoomAction(
            parent=self, plot=self._leftPanel.getPlotWidget()
        )
        toolbar.addAction(resetZoom)

        colormapAction = actions.control.ColormapAction(
            parent=self, plot=self._leftPanel.getPlotWidget()
        )
        colormapDialog = ColormapDialog(parent=self)

        colormapAction.setColorDialog(colormapDialog)
        toolbar.addAction(colormapAction)
        self._rightPanel.getPlotWidget().setDefaultColormap(
            self._leftPanel.getPlotWidget().getDefaultColormap()
        )

        toggleColorbarAction = actions.control.ColorBarAction(
            plot=self._leftPanel, parent=self
        )
        toggleColorbarAction.setCheckable(True)
        toggleColorbarAction.setChecked(False)
        toolbar.addAction(toggleColorbarAction)
        toggleColorbarAction.triggered.connect(self._toggleBothColorbars)

        toolbar.addSeparator()

        self._togglePeaksAction = qt.QAction(
            qtawesome.icon("fa6s.xmark", rotated=45), "Show/Hide Segmented Peaks", self
        )
        self._togglePeaksAction.setIconText("Peaks")
        self._togglePeaksAction.setCheckable(True)
        self._togglePeaksAction.setChecked(True)
        self._togglePeaksAction.triggered.connect(self._toggleSegmentPeaks)
        toolbar.addAction(self._togglePeaksAction)

        self._peakColorComboBox = PeakColorComboBox(
            parent=self, plot=self._rightPanel.getPlotWidget()
        )
        self._peakColorComboBox.currentColorChanged.connect(self._changePeakColor)
        toolbar.addWidget(self._peakColorComboBox)

    def _syncInteractionMode(self):
        mode = self._leftPanel.getPlotWidget().getInteractiveMode()
        self._rightPanel.getPlotWidget().setInteractiveMode(**mode)

    def _toggleBothColorbars(self, checked):
        self._leftPanel.setColorBarVisible(checked)
        self._rightPanel.setColorBarVisible(checked)

    def _syncAxes(self):
        self._xSync = SyncAxes(
            [
                self._leftPanel.getPlotWidget().getXAxis(),
                self._rightPanel.getPlotWidget().getXAxis(),
            ]
        )
        self._ySync = SyncAxes(
            [
                self._leftPanel.getPlotWidget().getYAxis(),
                self._rightPanel.getPlotWidget().getYAxis(),
            ]
        )

    def _toggleSegmentPeaks(self):
        scatter = self._rightPanel.getScatter()
        if scatter:
            scatter.setVisible(self._togglePeaksAction.isChecked())

    def _changePeakColor(self, newColor: str):
        self._rightPanel.setScatterColor(newColor)

    def setLeftImage(
        self, array2D: np.ndarray, title: str, xLabel: str, yLabel: str
    ) -> None:
        self._rightPanel.setImage(array2D, title, xLabel, yLabel)

    def setRightImage(
        self, array2D: np.ndarray, title: str, xLabel: str, yLabel: str
    ) -> None:
        self._leftPanel.setImage(array2D, title, xLabel, yLabel)

    def setRightScatter(self, x: np.ndarray, y: np.ndarray):
        self._rightPanel.setScatter(
            x=x,
            y=y,
            colormap={"colors": [rgba(self._peakColorComboBox.getCurrentColor())]},
            symbol="x",
        )
        # Sync the visibily of the scatter with the action status
        self._toggleSegmentPeaks()



from __future__ import annotations

from pathlib import Path

from ImageD11.unitcell import unitcell as UnitCell
from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.widgets.FloatEdit import FloatEdit
from silx.io.utils import DataUrl

from ewoks3dxrd.io import read_lattice_cell_data
from ewoks3dxrd.nexus.peaks import read_peaks_attributes
from ewoks3dxrd.nexus.utils import group_exists
from ewoks3dxrd.tasks.filter_by_index import FilterByIndexer, Inputs

from .common.ewoks3dxrd_widget import Ewoks3DXRDWidget
from .common.master_file_widget import MasterFileWidget
from .common.peak_filter_plot2d import PeakFilterPlot2D
from .common.process_name_line_edit import ProcessNameLineEdit
from .common.tuple_input_widget import TupleInputWidget


class OWIndexerFilter(Ewoks3DXRDWidget, ewokstaskclass=FilterByIndexer):
    name = "Index Filter"
    description = "Choose rings aligned peaks."
    icon = "icons/rings-smallest.svg"

    _ewoks_inputs_to_hide_from_orange = (
        "overwrite",
        "rings",
        "process_group_name",
    )

    def __init__(self):
        super().__init__()

        controlWidget = qt.QWidget()
        layout = qt.QVBoxLayout(controlWidget)

        settingsWidget = qt.QGroupBox("Index filter settings")
        settingsLayout = qt.QFormLayout(settingsWidget)
        self._ringIndexTuple = TupleInputWidget()
        settingsLayout.addRow("Select Rings", self._ringIndexTuple)
        self._lattice_file = MasterFileWidget("Select Lattice File")
        settingsLayout.addRow("Lattice file", self._lattice_file)
        self._ds_tol = FloatEdit()
        self._ds_tol.setAlignment(qt.Qt.AlignmentFlag.AlignLeft)
        settingsLayout.addRow("Reciprocal distance tolerance", self._ds_tol)
        layout.addWidget(settingsWidget)

        executeOverWriteLayout = qt.QFormLayout()
        self._inputPeaksUrl = qt.QLineEdit()
        self._processNameLineEdit = ProcessNameLineEdit(
            default=Inputs.model_fields["process_group_name"].default
        )
        self._overwrite = qt.QCheckBox()
        self._filterPeaksBtn = qt.QPushButton("Filter according to selected rings")

        executeOverWriteLayout.addRow("Incoming Data URL", self._inputPeaksUrl)
        executeOverWriteLayout.addRow(
            self._processNameLineEdit.label, self._processNameLineEdit
        )
        executeOverWriteLayout.addRow("Overwrite", self._overwrite)
        executeOverWriteLayout.addRow(self._filterPeaksBtn)
        layout.addLayout(executeOverWriteLayout)
        self._filterPeaksBtn.clicked.connect(self.execute_ewoks_task)

        self.addControlWidget(controlWidget)
        self._plot = PeakFilterPlot2D(self)
        self.addMainWidget(self._plot)

        self.registerInput(
            "intensity_filtered_data_url",
            self._getInputUrl,
            self._inputPeaksUrl.setText,
        )
        self.registerInput(
            "rings",
            self._ringIndexTuple.getValue,
            self._ringIndexTuple.setValue,
        )
        self.registerInput("ds_tol", self._ds_tol.value, self._ds_tol.setValue)
        self.registerInput(
            "lattice_file", self._lattice_file.getText, self._lattice_file.setText
        )
        self.registerInput(
            "process_group_name",
            self._processNameLineEdit.getText,
            self._processNameLineEdit.setText,
        )

        self.registerInput(
            "overwrite", self._overwrite.isChecked, self._overwrite.setChecked
        )
        self._restoreDefaultInputs()
        self._plotIncomingPeaks()

    def handleNewSignals(self):
        input_peaks_url = self.get_task_input_value("intensity_filtered_data_url", None)
        if input_peaks_url:
            self._inputPeaksUrl.setText(input_peaks_url)
        wavelength = self.get_task_input_value("wavelength", None)
        if wavelength:
            self.set_default_input("wavelength", wavelength)
        ds_tol = self.get_task_input_value("ds_tol", None)
        if ds_tol:
            self._ds_tol.setText(str(ds_tol))
        lattice_parameter_file = self.get_task_input_value("lattice_file", None)
        if lattice_parameter_file:
            self._lattice_file.setText(str(lattice_parameter_file))

        if (
            wavelength is None
            or ds_tol is None
            or lattice_parameter_file is None
            or input_peaks_url is None
        ):
            return
        self._validateInputs()
        self._plotIncomingPeaks()

    def _plotIncomingPeaks(self):
        input_url = self.get_task_input_value("intensity_filtered_data_url", None)
        if input_url is None:
            return
        data_url = DataUrl(input_url)
        nexus_file_path, lattice_data_group_path = (
            data_url.file_path(),
            data_url.data_path(),
        )
        intensity_3d_peaks = read_peaks_attributes(
            filename=nexus_file_path,
            process_group=lattice_data_group_path,
        )
        self._plot.addUnfilteredScatter(
            x=intensity_3d_peaks["ds"],
            y=intensity_3d_peaks["sum_intensity"],
            value=len(intensity_3d_peaks["sum_intensity"]) * [1],
            colormap=Colormap(name="viridis", autoscaleMode="percentile_1_99"),
        )
        self._plot.setGraphXLabel("Reciprocal distance (ds)")
        self._plot.setGraphYLabel("Sum Intensity")
        self._plot.setYAxisLogarithmic(True)
        self._plot.resetZoom()
        self._plotRings(
            ds_limit=max(intensity_3d_peaks["ds"]),
            y_limits=[
                min(intensity_3d_peaks["sum_intensity"]),
                max(intensity_3d_peaks["sum_intensity"]),
            ],
        )

    def _plotRings(self, ds_limit: float, y_limits):
        lattice_file_name = self.get_task_input_value("lattice_file")
        if not lattice_file_name:
            return
        unit_cell = read_lattice_cell_data(lattice_data_file_path=lattice_file_name)
        try:
            unit_cell = UnitCell(
                lattice_parameters=unit_cell.lattice_parameters,
                symmetry=int(unit_cell.space_group),
            )
            unit_cell.makerings(limit=ds_limit)

            for i, ring_ds in enumerate(unit_cell.ringds):
                self._plot.addCurve(
                    [ring_ds, ring_ds], y_limits, legend=f"Ring {i}", linewidth=2
                )
        except Exception as e:
            self.showError(e, title="Lattice Ring Plot")

    def handleSuccessfulExecution(self):
        if self.get_task_output_value("indexer_filtered_data_url", None) is None:
            return

        data_url = DataUrl(self.get_task_output_value("indexer_filtered_data_url"))
        nexus_file_path, data_group_path = (
            data_url.file_path(),
            data_url.data_path(),
        )
        index_3d_peaks = read_peaks_attributes(
            filename=nexus_file_path,
            process_group=data_group_path,
        )
        self._plot.addFilteredScatter(
            x=index_3d_peaks["ds"],
            y=index_3d_peaks["sum_intensity"],
            value=index_3d_peaks["Number_of_pixels"],
            colormap=Colormap(
                name="viridis", normalization="log", autoscaleMode="percentile_1_99"
            ),
        )
        self._plot.resetZoom()

    def _validateInputs(self):
        assert isinstance(float(self._ds_tol.text()), float)
        assert Path(str(self._lattice_file.getText().strip())).exists()
        input_peaks_url = DataUrl(self._inputPeaksUrl.text().strip())
        group_exists(
            filename=input_peaks_url.file_path(),
            data_group_path=input_peaks_url.data_path(),
        )

    def _getInputUrl(self) -> str:
        input_peaks_url = DataUrl(self._inputPeaksUrl.text().strip())
        group_exists(
            filename=input_peaks_url.file_path(),
            data_group_path=input_peaks_url.data_path(),
        )
        return input_peaks_url.path()

    def _getRingIndices(self):
        ringIndices = self._ringIndexTuple.getValue()
        if not ringIndices:
            raise ValueError("No ring selected")

        return ringIndices



from pathlib import Path
from typing import Optional

import h5py
from ewokscore import TaskWithProgress
from ImageD11 import frelon_peaksearch

from ..io import extract_sample_info, get_monitor_scale_factor
from ..models import (
    InputsWithOverwrite,
    SegmenterConfig,
    SegmenterCorrectionFiles,
    SegmenterFolderConfig,
)
from ..nexus.peaks import save_nexus_process
from ..nexus.utils import group_exists
from ..tqdm_progress_callback import TqdmProgressCallback


class Inputs(InputsWithOverwrite):
    folder_config: SegmenterFolderConfig
    segmenter_algo_params: SegmenterConfig
    correction_files: SegmenterCorrectionFiles
    monitor_name: Optional[str] = None


class SegmentScan(
    TaskWithProgress,
    input_model=Inputs,
    output_names=["sample_folder_info", "segmented_peaks_url"],
):
    """
    This task segments an entire scan folder,
    merges the peaks, and produces a 3D column file.
    The resulting 3D column peak file is saved.

    Outputs:

    - `sample_folder_info`: A Config information about raw scan sample
    - `segmented_peaks_url`: A Nexus file data url path to `segmented_3d_peaks` data
    """

    def run(self):
        inputs = Inputs(**self.get_input_values())
        seg_folder_config = inputs.folder_config

        detector = seg_folder_config.detector
        omega_motor = seg_folder_config.omega_motor
        master_file = seg_folder_config.master_file
        scan_number = seg_folder_config.scan_number

        _, sample_name, dset_name = extract_sample_info(master_file=master_file)

        masterfile_path = Path(master_file)
        if not masterfile_path.exists():
            raise FileNotFoundError(f"""{masterfile_path} does not exist.""")

        analyse_folder = seg_folder_config.analyse_folder
        output_folder = (
            Path(analyse_folder) / sample_name / f"{sample_name}_{dset_name}"
        )
        output_folder.mkdir(parents=True, exist_ok=True)
        nexus_file_path = output_folder / f"{sample_name}_{dset_name}.h5"
        seg_data_group_path = f"{str(scan_number)}.1/segmented_3d_peaks"
        overwrite = inputs.overwrite
        if not overwrite and group_exists(
            filename=nexus_file_path, data_group_path=seg_data_group_path
        ):
            raise ValueError(
                f"""Data group '{seg_data_group_path}' already exists in {nexus_file_path},
                Set `overwrite` to True if you wish to overwrite the existing data group.
                """
            )

        with h5py.File(masterfile_path, "r") as hin:
            omega_angles = hin[str(scan_number) + ".1"]["measurement"].get(
                omega_motor, None
            )
            omega_array = omega_angles[()]

        segmenter_cfg = inputs.segmenter_algo_params
        correction_files_config = inputs.correction_files
        segmenter_settings = {
            "bgfile": correction_files_config.bg_file,
            "maskfile": correction_files_config.mask_file,
            "darkfile": correction_files_config.dark_file,
            "flatfile": correction_files_config.flat_file,
            "threshold": segmenter_cfg.threshold,
            "smoothsigma": segmenter_cfg.smooth_sigma,
            "bgc": segmenter_cfg.bgc,
            "minpx": segmenter_cfg.min_px,
            "m_offset_thresh": segmenter_cfg.offset_threshold,
            "m_ratio_thresh": segmenter_cfg.ratio_threshold,
        }

        monitor_name = inputs.monitor_name
        scale_factor = (
            None
            if monitor_name is None
            else get_monitor_scale_factor(masterfile_path, scan_number, monitor_name)
        )

        all_frames_2d_peaks_list = frelon_peaksearch.segment_master_file(
            str(masterfile_path),
            str(scan_number) + ".1" + "/measurement/" + detector,
            omega_array,
            segmenter_settings,
            scale_factor=scale_factor,
            tqdm_class=TqdmProgressCallback,
            TaskInstance=self,
        )
        peaks_2d_dict, num_peaks = frelon_peaksearch.peaks_list_to_dict(
            all_frames_2d_peaks_list
        )
        # 3d merge from 2d peaks dict
        peak_3d_dict = frelon_peaksearch.do3dmerge(
            peaks_2d_dict, num_peaks, omega_array
        )

        self.outputs.sample_folder_info = {
            "omega_motor": omega_motor,
            "master_file": master_file,
            "scan_number": scan_number,
        }

        nxprocess_url = save_nexus_process(
            filename=nexus_file_path,
            entry_name=f"{str(scan_number)}.1",
            process_name="segmented_3d_peaks",
            peaks_data=peak_3d_dict,
            config_settings={
                "FolderFileSettings": seg_folder_config.model_dump(),
                "Segmenter_settings": segmenter_settings,
            },
            overwrite=overwrite,
        )
        self.outputs.segmented_peaks_url = nxprocess_url



from __future__ import annotations
from silx.gui import qt
import os
from ..common.filename_completer_line_edit import FilenameCompleterLineEdit


class MasterFileWidget(qt.QWidget):
    sigMasterFileChanged = qt.Signal(str)

    def __init__(
        self,
        dialogTitle="Select Master File",
        parent: qt.QWidget | None = None,
        **kwargs,
    ) -> None:
        super().__init__(parent=parent, **kwargs)

        self._line_edit = FilenameCompleterLineEdit(self)
        self._browse_btn = qt.QPushButton("Browse...")

        layout = qt.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._line_edit)
        layout.addWidget(self._browse_btn)

        self._dialog_title = dialogTitle
        self._browse_btn.clicked.connect(self._open_file_dialog)
        self._line_edit.textChanged.connect(self._on_text_changed)
        self._last_emitted_path = None

    def _open_file_dialog(self):
        path, _ = qt.QFileDialog.getOpenFileName(self, self._dialog_title)
        if path:
            self._line_edit.setText(path)

    def getText(self) -> str:
        return self._line_edit.text().strip()

    def setText(self, file_path):
        self._line_edit.setText(file_path)

    def _on_text_changed(self, text: str):
        path = text.strip()
        if os.path.isfile(path) and path != self._last_emitted_path:
            self._last_emitted_path = path
            self.sigMasterFileChanged.emit(path)



from __future__ import annotations
import os
from typing import Literal
import numpy as np
import h5py
from silx.gui import qt


def get_omega_array(
    file_path: str,
    omega_motor: str,
    scan_id: str,
) -> np.ndarray:
    with h5py.File(file_path, "r") as hin:
        omega_angles = hin[f"{scan_id}/measurement/{omega_motor}"]
        omega_array = omega_angles[()]
        return omega_array


def get_unique_instrument_keys(master_file: str, groups: list[str]) -> list[str]:
    unique_keys = set()
    with h5py.File(master_file) as hin:
        for group in groups:
            instrument_path = f"{group}/instrument"
            if instrument_path in hin:
                group_keys = hin[instrument_path].keys()
                unique_keys.update(group_keys)

    return sorted(unique_keys)


def find_possible_scan_numbers(master_file: str) -> list[int]:
    parent_dir = os.path.dirname(master_file)
    candidates = []
    for d in os.listdir(parent_dir):
        if d.lower().startswith("scan") and os.path.isdir(os.path.join(parent_dir, d)):
            digits = "".join(filter(str.isdigit, d))
            if digits:
                candidates.append(int(digits))
    return candidates


def ask_confirmation_to_repeat_segmentation() -> Literal["continue", "show", "cancel"]:
    """Asks the user what to do when segmentation parameters have not changed
    Returns:
    str: One of "continue", "show", or "cancel"
    """
    msg_box = qt.QMessageBox()
    msg_box.setIcon(qt.QMessageBox.Warning)
    msg_box.setWindowTitle("Segmentation Already Run")
    msg_box.setText("You've already run a full segmentation with these parameters.")
    msg_box.setInformativeText("What would you like to do?")

    repeat_btn = msg_box.addButton("Repeat anyway", qt.QMessageBox.AcceptRole)
    show_btn = msg_box.addButton("Show previous results", qt.QMessageBox.ActionRole)
    cancel_btn = msg_box.addButton("Cancel", qt.QMessageBox.RejectRole)
    msg_box.exec()

    clicked = msg_box.clickedButton()
    if clicked == repeat_btn:
        return "continue"
    elif clicked == show_btn:
        return "show"
    elif clicked == cancel_btn:
        return "cancel"



from ewoks import Task
from pydantic import Field

from ..models import DetectorCorrectionFiles, InputsWithOverwrite
from ..nexus.peaks import read_peaks_attributes, save_nexus_process
from ..nexus.utils import (
    check_throw_write_data_group_url,
    get_data_url_paths,
    get_entry_name,
    group_exists,
)
from ..utils import do_spatial_correction


class Inputs(InputsWithOverwrite):
    segmented_peaks_url: str = Field(
        description="Path to the NXprocess group containing the segmeted peaks"
    )
    correction_files: DetectorCorrectionFiles = Field(
        description="""two corrections are possible:

        - Spline correction: `correction_files` should be a string containing the path to the spline file
        - e2dx,e2dy correction: `correction_files` should be a tuple of 2 strings, the first one being the path to e2dx file, the second the path to the e2dy file
        - any other type will be treated as invalid input""",
        examples=["eiger.spline", ("eiger.e2dx", "eiger.e2dy")],
    )


class DetectorSpatialCorrection(
    Task,
    input_model=Inputs,
    output_names=["spatial_corrected_data_url"],
):
    """
    Does the detector spatial correction on the segmented 3d peaks and saves the corrected 3D column peak file

    Outputs:
    - `spatial_corrected_data_url`: A Nexus file path along with data entry point to `spatial_corrected_peaks` data group
    """

    def run(self):
        inputs = Inputs(**self.get_input_values())

        nexus_file_path, segmented_data_group_path = get_data_url_paths(
            inputs.segmented_peaks_url
        )

        if not group_exists(nexus_file_path, segmented_data_group_path):
            raise FileNotFoundError(
                f""" File or Data Group not Found Error,
                Either it is missing nexus file path: {nexus_file_path} or
                segmented data group path: {segmented_data_group_path}.
                """
            )

        output_entry = get_entry_name(segmented_data_group_path)
        output_process = "spatial_corrected_peaks"
        check_throw_write_data_group_url(
            overwrite=inputs.overwrite,
            filename=nexus_file_path,
            data_group_path=f"{output_entry}/{output_process}",
        )

        segmented_3d_peaks = read_peaks_attributes(
            filename=nexus_file_path,
            process_group=segmented_data_group_path,
        )
        columnfile_3d = do_spatial_correction(
            segmented_3d_peaks, inputs.correction_files
        )

        nxprocess_url = save_nexus_process(
            filename=nexus_file_path,
            entry_name=output_entry,
            process_name=output_process,
            peaks_data={"sc": columnfile_3d["sc"], "fc": columnfile_3d["fc"]},
            config_settings={
                "correction_files": inputs.correction_files,
                "in_peaks_data_from": inputs.segmented_peaks_url,
            },
            pks_axes=("fc", "sc"),
            overwrite=inputs.overwrite,
            ln_pks_from_group_name=segmented_data_group_path,
        )
        self.outputs.spatial_corrected_data_url = nxprocess_url



from ewoks3dxrd.tasks.make_grain_map import MakeGrainMap

from .conftest import assert_grain_map_results


def test_map_grains(tmp_path, nexus_3dxrd_file):
    inputs = {
        "hkl_tols": (0.05, 0.025, 0.01),
        "minpks": 120,
        "folder_file_config": {
            "omega_motor": "diffrz",
            "master_file": "/data/projects/id03_3dxrd/expt/RAW_DATA/FeAu_0p5_tR/FeAu_0p5_tR_ff1/FeAu_0p5_tR_ff1.h5",
            "scan_number": 1,
        },
        "intensity_fine_filtered_data_url": f"{nexus_3dxrd_file}::/1.1/intensity_filtered_inner_peaks",
        "indexed_grain_data_url": f"{nexus_3dxrd_file}::/1.1/indexed_ubi",
        "intensity_filtered_data_url": f"{nexus_3dxrd_file}::/1.1/intensity_filtered_all_peaks",
        "analyse_folder": tmp_path,
        "overwrite": True,
        "lattice_file": "/data/projects/id03_3dxrd/expt/PROCESSED_DATA/Fe.par",
    }

    task = MakeGrainMap(inputs=inputs)
    task.execute()

    assert_grain_map_results(task.outputs)



from ewoks3dxrd.tasks.grid_index_grains import GridIndexGrains

from ..nexus.utils import get_data_url_paths, group_exists


def test_grid_index_grains(tmp_path, nexus_3dxrd_file, id03_inp_config):
    inputs = {
        "indexer_filtered_data_url": f"{nexus_3dxrd_file}::/1.1/intensity_filtered_inner_peaks",
        "grid_index_parameters": {"NPKS": 3},
        "analyse_folder": tmp_path,
        "overwrite": True,
        "grid_step": 20,
        "grid_abs_x_limit": 10,
        "grid_abs_y_limit": 0,
        "grid_abs_z_limit": 0,
        "sample_config": id03_inp_config,
    }

    task = GridIndexGrains(inputs=inputs)
    task.execute()
    nexus_file, nexus_group = get_data_url_paths(
        task.outputs["grid_indexed_grain_data_url"]
    )
    assert group_exists(filename=nexus_file, data_group_path=nexus_group)



from __future__ import annotations

from pathlib import Path
from typing import Any

from silx.gui import qt

from ewoks3dxrd.io import read_lattice_cell_data
from ewoks3dxrd.models import UnitCellParameters

from ..common.collapsible_widget import CollapsibleWidget
from ..common.master_file_widget import MasterFileWidget


class LatticeParameters(qt.QGroupBox):
    sigLatticeParamsChanged = qt.Signal()

    def __init__(
        self,
        title: str = "Lattice Parameter Settings",
        parent: qt.QWidget | None = None,
        **kwargs,
    ) -> None:
        super().__init__(title, parent=parent, **kwargs)
        self._latticeName = None

        layout = qt.QVBoxLayout(self)

        fileLayout = qt.QFormLayout()
        self._latticeParFile = MasterFileWidget(dialogTitle="Lattice Parameter File")
        self._latticeParFile.sigMasterFileChanged.connect(self._onLatticeFileChanged)
        fileLayout.addRow("Lattice File:", self._latticeParFile)
        layout.addLayout(fileLayout)

        editingFieldsBox = CollapsibleWidget("Edit Parameters")
        formLayout = qt.QFormLayout()
        self._aEdit = qt.QDoubleSpinBox()
        self._aEdit.setSuffix(" Å")
        self._aEdit.setRange(0.01, 1000)
        self._aEdit.setValue(1)
        self._aEdit.setDecimals(4)
        formLayout.addRow("a:", self._aEdit)

        self._bEdit = qt.QDoubleSpinBox()
        self._bEdit.setSuffix(" Å")
        self._bEdit.setRange(0.01, 1000)
        self._bEdit.setValue(1)
        self._bEdit.setDecimals(4)
        formLayout.addRow("b:", self._bEdit)

        self._cEdit = qt.QDoubleSpinBox()
        self._cEdit.setSuffix(" Å")
        self._cEdit.setRange(0.01, 1000)
        self._cEdit.setValue(1)
        self._cEdit.setDecimals(4)
        formLayout.addRow("c:", self._cEdit)

        self._alphaEdit = qt.QDoubleSpinBox()
        self._alphaEdit.setSuffix(" °")
        self._alphaEdit.setRange(0, 179.999)
        self._alphaEdit.setValue(90.0)
        formLayout.addRow("α:", self._alphaEdit)

        self._betaEdit = qt.QDoubleSpinBox()
        self._betaEdit.setSuffix(" °")
        self._betaEdit.setRange(0, 179.999)
        self._betaEdit.setValue(90.0)
        formLayout.addRow("β:", self._betaEdit)

        self._gammaEdit = qt.QDoubleSpinBox()
        self._gammaEdit.setSuffix(" °")
        self._gammaEdit.setRange(0, 179.999)
        self._gammaEdit.setValue(90.0)
        formLayout.addRow("γ:", self._gammaEdit)

        self._spaceGroupEdit = qt.QSpinBox()
        self._spaceGroupEdit.setRange(1, 230)
        self._spaceGroupEdit.setValue(229)
        formLayout.addRow("Space Group (#):", self._spaceGroupEdit)

        editingFieldsBox.setCollapsed(True)
        editingFieldsBox.setLayout(formLayout)
        layout.addWidget(editingFieldsBox)
        layout.addStretch(1)

        for widget in [
            self._aEdit,
            self._bEdit,
            self._cEdit,
            self._alphaEdit,
            self._betaEdit,
            self._gammaEdit,
            self._spaceGroupEdit,
        ]:
            widget.valueChanged.connect(self.sigLatticeParamsChanged.emit)

    def _onLatticeFileChanged(self):
        filePath = Path(self._latticeParFile.getText())
        ext = filePath.suffix
        if ext == ".cif" or ext == ".par":
            unit_cell_parameters: UnitCellParameters = read_lattice_cell_data(filePath)
        else:
            qt.QMessageBox.warning(
                self, "Unsupported Format", "Only .cif and .par formats are supported."
            )
            return

        self._latticeName = filePath.stem

        self._aEdit.setValue(unit_cell_parameters.a)
        self._bEdit.setValue(unit_cell_parameters.b)
        self._cEdit.setValue(unit_cell_parameters.c)
        self._alphaEdit.setValue(unit_cell_parameters.alpha)
        self._betaEdit.setValue(unit_cell_parameters.beta)
        self._gammaEdit.setValue(unit_cell_parameters.gamma)
        self._spaceGroupEdit.setValue(int(unit_cell_parameters.space_group))

    def getLatticeParameters(self) -> dict[str, Any]:
        return {
            "lattice_parameters": (
                self._aEdit.value(),
                self._bEdit.value(),
                self._cEdit.value(),
                self._alphaEdit.value(),
                self._betaEdit.value(),
                self._gammaEdit.value(),
            ),
            "lattice_space_group": self._spaceGroupEdit.value(),
            "lattice_name": self._latticeName,
        }

    def setFile(self, value: str):
        self._latticeParFile.setText(value)



from __future__ import annotations

from pathlib import Path
from typing import List

from ImageD11.grain import grain as Grain
from ImageD11.grain import read_grain_file
from ImageD11.grid_index_parallel import grid_index_parallel

from .models import GridIndexParameters, Translations
from .nexus.utils import get_data_url_paths, get_entry_name
from .nexus.parameters import find_lattice_nexus_group_url
from .nexus.peaks import save_column_file_as_ascii

from .tmp_files import tmp_lattice_and_geo_file


def run_grid_indexing(
    input_data_url: str,
    grid_index_parameters: GridIndexParameters,
    analyse_folder: str | Path,
    translations: Translations,
) -> List[Grain]:
    nexus_file_path, indexer_filtered_data_url = get_data_url_paths(input_data_url)
    entry_name = get_entry_name(indexer_filtered_data_url)

    analyse_folder = (
        Path(analyse_folder) if analyse_folder else Path(nexus_file_path).parent
    )
    output_file = str(analyse_folder / "grid_indexed_grains.map")
    with tmp_lattice_and_geo_file(
        lat_par_data_url=find_lattice_nexus_group_url(input_data_url),
        geo_par_data_url=f"{nexus_file_path}::{entry_name}/geometry_updated_peaks",
    ) as par_file:
        grid_peaks_path = analyse_folder / "phase_index_filtered_peaks.flt"
        save_column_file_as_ascii(input_data_url, grid_peaks_path)
        grid_index_parallel(
            fltfile=grid_peaks_path,
            parfile=str(par_file),
            tmp=str(par_file.parent),
            gridpars={
                "output_filename": output_file,
                **grid_index_parameters.model_dump(),
            },
            translations=translations,
        )

    return read_grain_file(output_file)



from __future__ import annotations
import numpy as np
from ImageD11 import frelon_peaksearch


def segment_frame(
    raw_image: np.ndarray, params: dict, scale_factor: float | None = None
) -> tuple[np.ndarray, np.ndarray]:
    image_worker = frelon_peaksearch.worker(
        bgfile=params["correction_files"]["bg_file"],
        maskfile=params["correction_files"]["mask_file"],
        darkfile=params["correction_files"]["dark_file"],
        flatfile=params["correction_files"]["flat_file"],
        threshold=params["segmenter_config"]["threshold"],
        smoothsigma=params["segmenter_config"]["smooth_sigma"],
        bgc=params["segmenter_config"]["bgc"],
        minpx=params["segmenter_config"]["min_px"],
        m_offset_thresh=params["segmenter_config"]["offset_threshold"],
        m_ratio_thresh=params["segmenter_config"]["ratio_threshold"],
    )
    goodpeaks = image_worker.peaksearch(
        img=raw_image, omega=0, scale_factor=scale_factor
    )
    # 23 and 24 are the columns for fast_column index and slow column index
    peak_positions = goodpeaks[:, 23:25].T
    return image_worker.smoothed, peak_positions



from silx.gui import qt

from ..common.master_file_widget import MasterFileWidget
from .geometry_parameters import GeometryParameterGroupBox


class GeometryTransformationSettings(qt.QGroupBox):
    sigParametersChanged = qt.Signal()

    def __init__(self, parent=None):
        super().__init__("Geometry settings", parent)
        self.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding)

        layout = qt.QVBoxLayout()
        fileBtnLayout = qt.QFormLayout()
        self._geometryFileWidget = MasterFileWidget(
            dialogTitle="Geometry Parameter File"
        )
        fileBtnLayout.addRow("Import file", self._geometryFileWidget)
        layout.addLayout(fileBtnLayout)
        self._geoParameterGroup = GeometryParameterGroupBox(self)
        layout.addWidget(self._geoParameterGroup)
        self.setLayout(layout)

        self._geometryFileWidget.sigMasterFileChanged.connect(
            self._updateGeometryParameters
        )

    def _updateGeometryParameters(self, filePath: str):
        self._geoParameterGroup.fillGeometryValues(filePath=filePath)

    def getGeometryParameters(self) -> dict[str, str]:
        return self._geoParameterGroup.getGeometryParameters()

    def setGeometryFile(self, filePath: str):
        self._geometryFileWidget.setText(filePath)



from __future__ import annotations

from typing import Any

from silx.gui import qt

from ewoks3dxrd.models import SegmenterConfig

from ..common.collapsible_widget import CollapsibleWidget
from ..common.debounce_timer import DebounceTimer
from .constants import SEGMENTER_DEFAULTS, SEGMENTER_TOOLTIPS


class SegmenterParamGroupBox(CollapsibleWidget):
    sigParamsChanged = qt.Signal()

    def __init__(self, parent: qt.QWidget | None = None, **kwargs) -> None:
        super().__init__("Segmentation Parameters", parent=parent, **kwargs)
        seg_layout = qt.QFormLayout()
        self._threshold = qt.QLineEdit(str(SEGMENTER_DEFAULTS["threshold"]))
        self._smooth_sigma = qt.QLineEdit(str(SEGMENTER_DEFAULTS["smooth_sigma"]))
        self._bgc = qt.QLineEdit(str(SEGMENTER_DEFAULTS["bgc"]))
        self._min_px = qt.QLineEdit(str(SEGMENTER_DEFAULTS["min_px"]))
        self._offset_threshold = qt.QLineEdit(
            str(SEGMENTER_DEFAULTS["offset_threshold"])
        )
        self._ratio_threshold = qt.QLineEdit(str(SEGMENTER_DEFAULTS["ratio_threshold"]))

        self._threshold.setValidator(qt.QIntValidator())
        self._smooth_sigma.setValidator(qt.QDoubleValidator())
        self._bgc.setValidator(qt.QDoubleValidator())
        self._min_px.setValidator(qt.QIntValidator())
        self._offset_threshold.setValidator(qt.QIntValidator())
        self._ratio_threshold.setValidator(qt.QIntValidator())

        self._threshold.setToolTip(SEGMENTER_TOOLTIPS["threshold"])
        self._smooth_sigma.setToolTip(SEGMENTER_TOOLTIPS["smooth_sigma"])
        self._bgc.setToolTip(SEGMENTER_TOOLTIPS["bgc"])
        self._min_px.setToolTip(SEGMENTER_TOOLTIPS["min_px"])
        self._offset_threshold.setToolTip(SEGMENTER_TOOLTIPS["offset_threshold"])
        self._ratio_threshold.setToolTip(SEGMENTER_TOOLTIPS["ratio_threshold"])

        seg_layout.addRow("Threshold:", self._threshold)
        seg_layout.addRow("Smooth Sigma:", self._smooth_sigma)
        seg_layout.addRow("Background Constant:", self._bgc)
        seg_layout.addRow("Min Pixels:", self._min_px)
        seg_layout.addRow("Offset Threshold:", self._offset_threshold)
        seg_layout.addRow("Ratio Threshold:", self._ratio_threshold)
        self.setLayout(seg_layout)

        self._last_params: dict[str, Any] = {}
        self._debounce_timer = DebounceTimer(
            callback=self._on_param_changed, timeout_ms=200, parent=self
        )

        for widget in [
            self._threshold,
            self._smooth_sigma,
            self._bgc,
            self._min_px,
            self._offset_threshold,
            self._ratio_threshold,
        ]:
            widget.textChanged.connect(self._debounce_timer.start)

    def getConfig(self) -> SegmenterConfig:
        return SegmenterConfig(
            threshold=int(self._threshold.text()),
            smooth_sigma=self._to_locale_float(self._smooth_sigma),
            bgc=self._to_locale_float(self._bgc),
            min_px=int(self._min_px.text()),
            offset_threshold=int(self._offset_threshold.text()),
            ratio_threshold=int(self._ratio_threshold.text()),
        )

    def setConfig(self, config: SegmenterConfig):
        self._threshold.setText(str(config.threshold))
        self._smooth_sigma.setText(str(config.smooth_sigma))
        self._bgc.setText(str(config.bgc))
        self._offset_threshold.setText(str(config.offset_threshold))
        self._ratio_threshold.setText(str(config.ratio_threshold))

    def _to_locale_float(self, text_field: qt.QLineEdit):
        value, ok = self.locale().toFloat(text_field.text())
        if not ok:
            raise ValueError(f"Invalid float input: '{text_field.text()}'")
        return value

    def _on_param_changed(self):
        params = self.getConfig()
        if not params or self._last_params == params:
            return

        self._last_params = params.model_dump()
        self.sigParamsChanged.emit()



from __future__ import annotations

import os
from typing import Any

from silx.gui import qt

from ..common.collapsible_widget import CollapsibleWidget
from ..common.debounce_timer import DebounceTimer
from ..common.file_folder_browse_button import FileFolderBrowseButton
from .constants import DEFAULTS_CORRECTION_MAP, SEGMENTER_TOOLTIPS, CorrectionFiles


class CorrectionGroupBox(CollapsibleWidget):
    sigParamsChanged = qt.Signal()

    def __init__(self, parent: qt.QWidget | None = None, **kwargs) -> None:
        super().__init__("Correction Files", parent=parent, **kwargs)

        corr_layout = qt.QFormLayout()
        self._bg_file_btn = FileFolderBrowseButton(dialogTitle="Select Background File")
        self._mask_file_btn = FileFolderBrowseButton(dialogTitle="Select Mask File")
        self._dark_file_btn = FileFolderBrowseButton(dialogTitle="Select Dark File")
        self._flat_file_btn = FileFolderBrowseButton(
            dialogTitle="Select Flat Field File"
        )

        self._bg_file_btn.setToolTip(SEGMENTER_TOOLTIPS["bg_file"])
        self._mask_file_btn.setToolTip(SEGMENTER_TOOLTIPS["mask_file"])
        self._dark_file_btn.setToolTip(SEGMENTER_TOOLTIPS["dark_file"])
        self._flat_file_btn.setToolTip(SEGMENTER_TOOLTIPS["flat_file"])

        self._last_params: dict[str, str] = {}
        self._debounce_timer = DebounceTimer(
            callback=self._on_param_changed, timeout_ms=200, parent=self
        )

        for widget in [
            self._bg_file_btn,
            self._mask_file_btn,
            self._dark_file_btn,
            self._flat_file_btn,
        ]:
            widget._line_edit.textChanged.connect(self._debounce_timer.start)

        corr_layout.addRow("Background File:", self._bg_file_btn)
        corr_layout.addRow("Mask File:", self._mask_file_btn)
        corr_layout.addRow("Dark File:", self._dark_file_btn)
        corr_layout.addRow("Flat Field File:", self._flat_file_btn)

        self.setLayout(corr_layout)

    def setDefaultCorrectionFiles(self, motor_name: str) -> None:
        defaults = DEFAULTS_CORRECTION_MAP.get(motor_name)
        if not defaults:
            return
        self._bg_file_btn.clearText()
        self._mask_file_btn.clearText()
        self._dark_file_btn.clearText()
        self._flat_file_btn.clearText()

        self.setCorrectionFiles(defaults)

    def getCorrectionFiles(self) -> CorrectionFiles:
        return {
            "bg_file": self._bg_file_btn.getText() or None,
            "mask_file": self._mask_file_btn.getText() or None,
            "dark_file": self._dark_file_btn.getText() or None,
            "flat_file": self._flat_file_btn.getText() or None,
        }

    def setCorrectionFiles(self, correctionFiles: dict[str, Any]):
        bg_file = correctionFiles.get("bg_file")
        if bg_file:
            self._bg_file_btn.setText(bg_file)

        mask_file = correctionFiles.get("mask_file")
        if mask_file:
            self._mask_file_btn.setText(mask_file)

        dark_file = correctionFiles.get("dark_file")
        if dark_file:
            self._dark_file_btn.setText(dark_file)

        flat_file = correctionFiles.get("flat_file")
        if flat_file:
            self._flat_file_btn.setText(flat_file)

    def _on_param_changed(self):
        params = self.getCorrectionFiles()
        if not params or self._last_params == params:
            return

        invalid_files = {
            key: path
            for key, path in params.items()
            if path is not None and not os.path.isfile(path)
        }

        if invalid_files:
            """
            msg = "The following file(s) do not exist:\n"
            msg += "\n".join(f"- {key}: {path}" for key, path in invalid_files.items())
            qt.QMessageBox.critical(self, "Invalid Correction Files", msg)
            """
            return

        self._last_params = params
        self.sigParamsChanged.emit()



from __future__ import annotations
from pathlib import Path
from typing import Tuple
import h5py
from silx.io.utils import DataUrl
import posixpath


def group_exists(filename: str | Path, data_group_path: str) -> bool:
    if isinstance(filename, str):
        filename = Path(filename)
    if not filename.exists():
        return False

    with h5py.File(filename, "r") as file:
        if data_group_path in file:
            return True
    return False


def check_throw_write_data_group_url(
    overwrite: bool, filename: str | Path, data_group_path: str
):
    if not overwrite and group_exists(
        filename=filename, data_group_path=data_group_path
    ):
        raise ValueError(
            f"""Data group '{data_group_path}' already exists in {filename},
            Set `overwrite` to True if you wish to overwrite the existing data group.
            """
        )


def get_data_url_paths(data_url_as_str: str) -> Tuple[str, str]:
    data_url = DataUrl(data_url_as_str)
    file_path = data_url.file_path()
    data_path = data_url.data_path()
    if not Path(file_path).exists():
        raise FileNotFoundError(f"{file_path} not found")
    if data_path is None or not group_exists(file_path, data_path):
        raise KeyError(f"{data_path} not found in {file_path}")

    return file_path, data_path


def get_entry_name(process_group_path: str):
    return posixpath.dirname(process_group_path)



import h5py
import pytest
from silx.io.url import DataUrl

from ewoks3dxrd.tasks.filter_by_intensity import FilterByIntensity


def test_execution(nexus_3dxrd_file):
    inputs = {
        "lattice_filtered_data_url": f"{nexus_3dxrd_file}::/1.1/phase_filtered_inner_peaks",
        "intensity_frac": 0.8,
    }

    task = FilterByIntensity(inputs=inputs)
    task.execute()
    output_url = DataUrl(task.outputs["intensity_filtered_data_url"])

    with h5py.File(output_url.file_path(), "r") as h5file:
        process_group = h5file[output_url.data_path()]
        assert isinstance(process_group, h5py.Group)
        assert isinstance(process_group["peaks/Number_of_pixels"], h5py.Dataset)


def test_wrong_intensity_frac(nexus_3dxrd_file):
    inputs = {
        "lattice_filtered_data_url": f"{nexus_3dxrd_file}::/1.1/phase_filtered_inner_peaks",
        "intensity_frac": 5,
    }

    task = FilterByIntensity(inputs=inputs)
    with pytest.raises(RuntimeError) as e:
        task.execute()
    original_exc = e.value.__cause__
    assert "Input should be less than or equal to 1" in str(original_exc)



from __future__ import annotations

import os

from silx.gui import qt

from ewoks3dxrd.models import SegmenterFolderConfig

from ..common.file_folder_browse_button import FileFolderBrowseButton
from ..common.master_file_widget import MasterFileWidget
from .constants import Detector, OmegaMotor
from .utils import find_possible_scan_numbers, get_unique_instrument_keys


class FolderMetadataGroupBox(qt.QGroupBox):
    sigMasterFileChanged = qt.Signal(str)

    def __init__(self, parent: qt.QWidget | None = None, **kwargs) -> None:
        super().__init__("Folder and Metadata Settings", parent=parent, **kwargs)
        folder_layout = qt.QFormLayout(self)
        self._master_file_path = MasterFileWidget(
            dialogTitle="3DXRD Experiment Master File"
        )
        folder_layout.addRow("Master File:", self._master_file_path)
        self._master_file_path.sigMasterFileChanged.connect(
            self._on_master_file_changed
        )

        self._omega_motor = qt.QComboBox()
        self._omega_motor.addItems([e.value for e in OmegaMotor])
        folder_layout.addRow("Omega Motor:", self._omega_motor)

        self._detector = qt.QComboBox()
        self._detector.addItems([e.value for e in Detector])
        folder_layout.addRow("Detector:", self._detector)

        self._scan_number = qt.QComboBox()
        folder_layout.addRow("Scan Number", self._scan_number)

        self._analyse_folder_path = FileFolderBrowseButton(
            dialogTitle="Select Analysis Folder", directory=True
        )
        folder_layout.addRow("Analyse Folder Path:", self._analyse_folder_path)

    def _on_master_file_changed(self, master_file_path: str):
        self._set_default_folder_metadata(master_file_path)
        instrument_keys = get_unique_instrument_keys(
            master_file=master_file_path, groups=["1.1", "1.2"]
        )
        self._set_default_detector(instrument_keys)
        self._set_default_motor(instrument_keys)

        self.sigMasterFileChanged.emit(master_file_path)

    def _set_default_folder_metadata(self, master_file: str):
        parent_dir = os.path.dirname(master_file)
        self._scan_number.clear()
        candidates = find_possible_scan_numbers(master_file)
        for c in candidates:
            self._scan_number.addItem(str(c))
        self._scan_number.setDisabled(len(candidates) <= 1)

        processed_data_path = os.path.dirname(parent_dir)
        processed_data_path = os.path.dirname(processed_data_path)
        processed_data_path = os.path.dirname(processed_data_path)
        processed_data_path = os.path.join(processed_data_path, "PROCESSED_DATA")
        if os.path.exists(processed_data_path):
            self._analyse_folder_path.setText(processed_data_path)

    def _set_default_detector(self, instrument_keys):
        detector_match = None
        for detector in Detector:
            if detector in instrument_keys:
                detector_match = detector
                break

        if detector_match:
            idx = self._detector.findText(detector_match)
            if idx != -1:
                self._detector.setCurrentIndex(idx)
                self._detector.setEnabled(False)

    def _set_default_motor(self, instrument_keys):
        omega_match = None
        for motor in OmegaMotor:
            if motor in instrument_keys:
                omega_match = motor
                break

        if omega_match:
            idx = self._omega_motor.findText(omega_match)
            if idx != -1:
                self._omega_motor.setCurrentIndex(idx)
                self._omega_motor.setEnabled(False)

    def getConfig(self) -> SegmenterFolderConfig:
        return SegmenterFolderConfig(
            omega_motor=self._omega_motor.currentText(),
            master_file=self._master_file_path.getText(),
            scan_number=int(self._scan_number.currentText()),
            detector=self._detector.currentText(),
            analyse_folder=self._analyse_folder_path.getText() or None,
        )

    def setConfig(self, config: SegmenterFolderConfig):
        self._master_file_path.setText(config.master_file)
        self._scan_number.setCurrentText(str(config.scan_number))
        self._omega_motor.setCurrentText(config.omega_motor)
        self._detector.setCurrentText(config.detector)
        if config.analyse_folder:
            self._analyse_folder_path.setText(config.analyse_folder)

    def getScanNumber(self) -> int:
        return int(self._scan_number.currentText())

    def getMasterFilePath(self) -> str:
        return self._master_file_path.getText()



from __future__ import annotations

import os
from typing import Any

from silx.gui import qt

from ewoks3dxrd.models import DetectorCorrectionFiles

from ..common.file_folder_browse_button import FileFolderBrowseButton


class DetectorCorrectionSettings(qt.QWidget):
    def __init__(self, parent: qt.QWidget | None = None) -> None:
        super().__init__(parent=parent)

        container_layout = qt.QVBoxLayout(self)
        self._radio_spline = qt.QRadioButton("Use Spline File")
        self._radio_correction_files = qt.QRadioButton("Use DX/DY Correction Files")
        self._radio_spline.setChecked(True)

        vbox_spline = qt.QVBoxLayout()
        vbox_spline.addWidget(self._radio_spline)
        self._spline_file = FileFolderBrowseButton(dialogTitle="Spline File")
        vbox_spline.addWidget(self._spline_file)

        vbox_correction = qt.QVBoxLayout()
        vbox_correction.addWidget(self._radio_correction_files)
        self._x_file = FileFolderBrowseButton(dialogTitle="Dx file")
        self._y_file = FileFolderBrowseButton(dialogTitle="Dy file")
        vbox_correction.addWidget(self._x_file)
        vbox_correction.addWidget(self._y_file)

        container_layout.addLayout(vbox_spline)
        container_layout.addLayout(vbox_correction)

        self._radio_spline.toggled.connect(self._toggle_correction_mode)
        self._radio_correction_files.toggled.connect(self._toggle_correction_mode)
        self._toggle_correction_mode()

    def _toggle_correction_mode(self):
        use_spline = self._radio_spline.isChecked()
        self._spline_file.setEnabled(use_spline)
        self._x_file.setDisabled(use_spline)
        self._y_file.setDisabled(use_spline)

    def getCorrectionFiles(self) -> DetectorCorrectionFiles:
        if self._radio_spline.isChecked():
            correction_files = self._spline_file.getText()
        else:
            correction_files = (self._x_file.getText(), self._y_file.getText())

        if not all(correction_files):
            raise ValueError("Set detector correction file(s).")

        return self._validate_correction_files(correction_files)

    def setCorrectionFiles(self, files: Any):
        correction_files = self._validate_correction_files(files)
        if isinstance(correction_files, str):
            self._spline_file.setText(correction_files)
        else:
            x_file, y_file = correction_files
            self._x_file.setText(x_file)
            self._y_file.setText(y_file)

    def _validate_correction_files(self, files: Any) -> DetectorCorrectionFiles:
        if isinstance(files, str):
            if not os.path.exists(files):
                raise FileNotFoundError(f"File does not exist: {files}")

        elif isinstance(files, tuple) and len(files) == 2:
            for f in files:
                if not os.path.exists(f):
                    raise FileNotFoundError(f"File does not exist: {f}")
        else:
            raise ValueError(f"Given correction files are not valid: {files}")

        return files



from pathlib import Path

from ImageD11.unitcell import unitcell as UnitCell
from silx.gui import qt
from silx.gui.colors import Colormap
from silx.io.utils import DataUrl

from ewoks3dxrd.io import save_par_file
from ewoks3dxrd.nexus.peaks import read_peaks_attributes
from ewoks3dxrd.tasks.geometry_transformation import GeometryTransformation

from .common.ewoks3dxrd_plot2d import Ewoks3DXRDPlot2D
from .common.ewoks3dxrd_widget import Ewoks3DXRDWidget
from .geometry.geometry_transformation_settings import GeometryTransformationSettings
from .lattice.lattice_parameters import LatticeParameters


class OWGeometryTransformation(Ewoks3DXRDWidget, ewokstaskclass=GeometryTransformation):
    name = "Geometry Transformation"
    description = (
        "Generate ds, g-vectors, eta, etc on detector corrected segmented 3D Peaks."
    )
    icon = "icons/settings.svg"
    _ewoks_inputs_to_hide_from_orange = (
        "overwrite",
        "geometry_par_file",
    )

    def __init__(self):
        super().__init__()

        self._settingsPanelWidget = qt.QWidget(self)
        settingLayout = qt.QVBoxLayout(self._settingsPanelWidget)
        self._latticeGroup = LatticeParameters("Show lattice rings", self)
        self._latticeGroup.setCheckable(True)
        self._ds_ring_limits = 1.5
        self._latticeGroup.sigLatticeParamsChanged.connect(self._plotRings)
        self._latticeGroup.toggled.connect(self._plotRings)
        self._latticeGroup.toggled.connect(self._latticeGroup.layout().setEnabled)
        self._settingsPanel = GeometryTransformationSettings(self)

        scrollContent = qt.QWidget()
        scrollLayout = qt.QVBoxLayout(scrollContent)
        scrollLayout.addWidget(self._settingsPanel)
        scrollLayout.addWidget(self._latticeGroup)
        scrollLayout.addStretch(1)  # Optional: pushes widgets to top

        scrollArea = qt.QScrollArea(self)
        scrollArea.setWidgetResizable(True)
        scrollArea.setWidget(scrollContent)

        settingLayout.addWidget(scrollArea)

        executeOverWriteLayout = qt.QFormLayout()
        self._detector_corrected_url = qt.QLineEdit()
        self._overwrite = qt.QCheckBox()
        self._computeGeometryCalculationBtn = qt.QPushButton(
            "Compute Geometry Vectors for Peaks"
        )
        executeOverWriteLayout.addRow(
            "Input: Detector Corrected URL", self._detector_corrected_url
        )
        executeOverWriteLayout.addRow("Overwrite", self._overwrite)
        executeOverWriteLayout.addRow(self._computeGeometryCalculationBtn)
        settingLayout.addLayout(executeOverWriteLayout)
        self._computeGeometryCalculationBtn.clicked.connect(self.execute_ewoks_task)
        self.addControlWidget(self._settingsPanelWidget)

        self._plot = Ewoks3DXRDPlot2D(self)
        self._plot.setKeepDataAspectRatio(False)
        self._plot.getColorBarWidget().setVisible(False)
        self._plot.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding)
        self.addMainWidget(self._plot)
        self.registerInput(
            "spatial_corrected_data_url",
            self._validateInputUrl,
            self._detector_corrected_url.setText,
        )
        self.registerInput(
            "geometry_par_file",
            self._prepareGeometryFile,
            self._settingsPanel.setGeometryFile,
        )
        self.registerInput(
            "overwrite", self._overwrite.isChecked, self._overwrite.setChecked
        )
        self._restoreDefaultInputs()

    def _plotRings(self, limit: float | None = None):
        if not self._latticeGroup.isChecked():
            self._plot.clearCurves()
            return

        if limit is None:
            limit = self._ds_ring_limits
        else:
            self._ds_ring_limits = limit

        params = self._latticeGroup.getLatticeParameters()
        try:
            unit_cell = UnitCell(
                lattice_parameters=params["lattice_parameters"],
                symmetry=int(params["lattice_space_group"]),
            )
            unit_cell.makerings(limit=limit)

            for i, ring_ds in enumerate(unit_cell.ringds):
                self._plot.addCurve(
                    [ring_ds, ring_ds], [-190, 190], legend=f"Ring_{i}", linewidth=2
                )
            self._plot.setGraphXLabel("Reciprocal Distance (ds)")
            self._plot.setGraphYLabel("Azimuthal Angle (eta)")
        except Exception as e:
            self.showError(e)

    def handleNewSignals(self):
        spatial_corrected_data_url = self.get_task_input_value(
            "spatial_corrected_data_url"
        )
        if spatial_corrected_data_url is not None:
            self._detector_corrected_url.setText(spatial_corrected_data_url)

    def handleSuccessfulExecution(self):
        if self.get_task_output_value("geometry_updated_data_url") is None:
            return

        data_url = DataUrl(self.get_task_output_value("geometry_updated_data_url"))
        nexus_file_path, geometry_data_group_path = (
            data_url.file_path(),
            data_url.data_path(),
        )
        geo_3d_peaks = read_peaks_attributes(
            filename=nexus_file_path,
            process_group=geometry_data_group_path,
        )
        scatter = self._plot.addScatter(
            x=geo_3d_peaks["ds"],
            y=geo_3d_peaks["eta"],
            value=geo_3d_peaks["Number_of_pixels"],
            colormap=Colormap(
                name="viridis", normalization="log", autoscaleMode="percentile_1_99"
            ),
            symbol="o",
            legend="aligned Peaks",
        )
        scatter.setSymbolSize(7)
        self._plot.setGraphXLabel("Reciprocal distance (ds)")
        self._plot.setGraphYLabel("Azimuthal angle (eta)")
        self._plot.resetZoom()
        self._plotRings(limit=max(geo_3d_peaks["ds"]))

    def _validateInputUrl(self):
        input_url = self._detector_corrected_url.text().strip()
        if not input_url:
            raise ValueError("No detector corrected data URL to process.")

        return input_url

    def _prepareGeometryFile(self) -> str:
        file_path = Path(DataUrl(self._detector_corrected_url.text()).file_path())
        geometry_file = file_path.parent / "geometry_tdxrd.par"
        parameters = self._settingsPanel.getGeometryParameters()
        if not parameters:
            raise ValueError("No geometry parameters to write")
        save_par_file(geometry_file, parameters)
        return str(geometry_file)



from __future__ import annotations

import numpy as np
from silx.gui import qt
from silx.gui.plot import Plot1D

from ewoks3dxrd.io import get_monitor_data

from ..common.collapsible_widget import CollapsibleWidget
from .constants import MONITOR_KEYS
from .utils import get_unique_instrument_keys


class MonitorSegmentGroupBox(CollapsibleWidget):

    def __init__(self, parent: qt.QWidget | None = None, **kwargs) -> None:
        super().__init__("Monitor Settings", parent=parent, **kwargs)

        monitor_layout = qt.QFormLayout()
        self._monitor_name = qt.QComboBox()
        self._monitor_name.addItem("None")
        self._monitor_name.currentTextChanged.connect(self._toggle_monitor_plot_btn)
        self._monitor_plot_btn = qt.QPushButton("Show Monitor Data")
        self._monitor_plot_btn.clicked.connect(self._on_show_monitor_clicked)
        self._monitor_plot_btn.setDisabled(True)

        monitor_layout.addRow("Monitor:", self._monitor_name)
        monitor_layout.addRow(self._monitor_plot_btn)

        self._scan_number = None
        self.setLayout(monitor_layout)

    def _on_show_monitor_clicked(self):
        current_monitor = self._monitor_name.currentText()
        if current_monitor == "None" or self._scan_number is None:
            return
        y_data = get_monitor_data(
            masterfile_path=self._master_file_path,
            scan_number=self._scan_number,
            monitor_name=current_monitor,
        )
        plot = Plot1D()
        plot.setWindowTitle(f"Monitor Name: {current_monitor}")
        plot.addCurve(range(len(y_data)), y_data, legend=current_monitor)
        plot.resize(800, 600)
        plot.show()

    def fillWidgetValues(self, master_file_path: str, scan_number: str):
        self._scan_number = scan_number
        self._master_file_path = master_file_path
        instrument_keys = get_unique_instrument_keys(
            master_file=master_file_path,
            groups=[self._scan_number + ".1", self._scan_number + ".2"],
        )

        monitor_keys = [
            k for k in instrument_keys if any(k.startswith(m) for m in MONITOR_KEYS)
        ]
        self._monitor_name.clear()
        self._monitor_name.addItem("None")
        for m in monitor_keys:
            monitor_data = get_monitor_data(
                masterfile_path=self._master_file_path,
                scan_number=self._scan_number,
                monitor_name=m,
            )
            if np.any(monitor_data <= 0):
                continue
            else:
                self._monitor_name.addItem(m)

    def getMonitorName(self) -> str | None:
        if self._monitor_name.currentText() == "None":
            return None

        return self._monitor_name.currentText()

    def setMonitorName(self, monitorName: str | None):
        if monitorName is None:
            self._monitor_name.setCurrentText("None")

        self._monitor_name.setCurrentText(monitorName)

    def _toggle_monitor_plot_btn(self, text: str):
        self._monitor_plot_btn.setDisabled(text == "None")



from __future__ import annotations
from typing import Sequence, Tuple
from ImageD11 import columnfile as PeakColumnFile
from ImageD11.unitcell import unitcell as UnitCell
from pathlib import Path
from .models import UnitCellParameters
from ImageD11.peakselect import (
    sorted_peak_intensity_mask_and_cumsum,
    filter_peaks_by_phase,
)
from .io import read_lattice_cell_data
import numpy as np
from ImageD11.indexing import indexer_from_colfile_and_ucell


def filter_by_intensity(peaks: dict, intensity_frac: float) -> dict:
    cf = PeakColumnFile.colfile_from_dict(peaks)
    mask, _ = sorted_peak_intensity_mask_and_cumsum(colf=cf, frac=intensity_frac)
    cf.filter(mask)

    return {key: cf[key] for key in cf.keys()}


def filter_by_phase(
    peaks: dict,
    lattice_file: Path,
    ds_tol: float,
    ds_max: float | None = None,
) -> Tuple[dict, UnitCellParameters]:
    raw_cf = PeakColumnFile.colfile_from_dict(peaks)

    unit_cell_parameters = read_lattice_cell_data(lattice_file)
    unit_cell = UnitCell(
        lattice_parameters=unit_cell_parameters.lattice_parameters,
        symmetry=unit_cell_parameters.space_group,
    )
    unit_cell.makerings(limit=raw_cf.ds.max())

    if ds_max is None:
        ds_max = raw_cf.ds.max()
    cf = raw_cf.copyrows(raw_cf.ds <= ds_max)
    filtered_cf = filter_peaks_by_phase(
        cf=cf,
        dstol=ds_tol,
        dsmax=ds_max,
        cell=unit_cell,
    )

    return {key: filtered_cf[key] for key in filtered_cf.keys()}, unit_cell_parameters


def filter_with_indexer(
    peak_3d_dict: dict,
    unit_cell: UnitCell,
    wavelength: float,
    phase_ds_tolerance: float,
    rings: Sequence[int],
) -> dict:
    """Using indexer and sequence of Ring indices, filter the given 3d peaks"""
    peaks_cf = PeakColumnFile.colfile_from_dict(peak_3d_dict)
    indexer = indexer_from_colfile_and_ucell(
        colfile=peaks_cf,
        ucell=unit_cell,
        wavelength=wavelength,
        ds_tol=phase_ds_tolerance,
    )
    indexer.assigntorings()

    mask = np.zeros(peaks_cf.nrows, dtype=bool)
    for ring in rings:
        mask |= indexer.ra == ring
    peaks_cf.filter(mask)

    return {key: peaks_cf[key] for key in peaks_cf.keys()}



from __future__ import annotations

import numpy as np
from silx.gui import qt
from silx.gui.colors import Colormap
from silx.io.utils import DataUrl

from ewoks3dxrd.nexus.peaks import read_peaks_attributes
from ewoks3dxrd.tasks.detector_spatial_correction import DetectorSpatialCorrection

from .common.ewoks3dxrd_plot2d import Ewoks3DXRDPlot2D
from .common.ewoks3dxrd_widget import Ewoks3DXRDWidget
from .detector.detector_correction_settings import DetectorCorrectionSettings


class OWDetectorCorrection(Ewoks3DXRDWidget, ewokstaskclass=DetectorSpatialCorrection):
    name = "Detector Correction"
    description = "Correct spatial deformed detector into plane"
    icon = "icons/distortion_plane.svg"

    _ewoks_inputs_to_hide_from_orange = (
        "overwrite",
        "correction_files",
    )

    def __init__(self):
        super().__init__()

        settings_panel_widget = qt.QWidget(self)
        setting_layout = qt.QVBoxLayout(settings_panel_widget)
        self._settings_widget = DetectorCorrectionSettings()

        action_layout = qt.QFormLayout()
        self._segmenter_url = qt.QLineEdit()
        self._overwrite = qt.QCheckBox("Overwrite")
        execute_btn = qt.QPushButton("Execute Detector Correction")
        execute_btn.clicked.connect(self.execute_ewoks_task)
        action_layout.addRow("Segmented URL", self._segmenter_url)
        self._segmenter_url.textChanged.connect(self._drawInputSegmentedPeaks)
        action_layout.addWidget(self._overwrite)
        action_layout.addWidget(execute_btn)

        setting_layout.addWidget(self._settings_widget, stretch=1)
        setting_layout.addLayout(action_layout)
        self.addControlWidget(settings_panel_widget)

        self._plot = Ewoks3DXRDPlot2D(self)
        self._plot.getColorBarWidget().setVisible(False)
        self.addMainWidget(self._plot)
        self.registerInput(
            "segmented_peaks_url", self._validateInputs, self._segmenter_url.setText
        )
        self.registerInput(
            "correction_files",
            self._settings_widget.getCorrectionFiles,
            self._settings_widget.setCorrectionFiles,
        )
        self.registerInput(
            "overwrite", self._overwrite.isChecked, self._overwrite.setChecked
        )
        self._restoreDefaultInputs()

    def handleNewSignals(self):
        segmented_peaks_url = self.get_task_input_value("segmented_peaks_url", None)
        if segmented_peaks_url is not None:
            self._segmenter_url.setText(segmented_peaks_url)
            self._drawInputSegmentedPeaks()

    def _drawInputSegmentedPeaks(self):
        data_url = DataUrl(self._segmenter_url.text().strip())
        nexus_file_path = data_url.file_path()
        segmented_data_group_path = data_url.data_path()

        segmented_3d_peaks = read_peaks_attributes(
            filename=nexus_file_path,
            process_group=segmented_data_group_path,
        )
        scatter = self._plot.addScatter(
            x=segmented_3d_peaks["f_raw"],
            y=segmented_3d_peaks["s_raw"],
            value=np.ones(len(segmented_3d_peaks["f_raw"])),
            colormap=Colormap(
                colors=np.array(
                    [
                        [0.0, 0.0, 1.0],
                    ],
                    dtype=np.float32,
                )
            ),
            symbol="o",
            legend="rawPixels",
        )
        scatter.setSymbolSize(7)
        self._plot.setGraphXLabel("fast column")
        self._plot.setGraphYLabel("slow column")
        self._plot.setGraphTitle("Raw Pixels O (blue)")
        self._plot.resetZoom()

    def handleSuccessfulExecution(self):
        data_url = DataUrl(self.get_task_output_value("spatial_corrected_data_url"))
        nexus_file_path = data_url.file_path()
        detector_data_group_path = data_url.data_path()

        det_corrected_3d_peaks = read_peaks_attributes(
            filename=nexus_file_path,
            process_group=detector_data_group_path,
        )
        scatter = self._plot.addScatter(
            x=det_corrected_3d_peaks["fc"],
            y=det_corrected_3d_peaks["sc"],
            value=np.ones(len(det_corrected_3d_peaks["fc"])),
            colormap=Colormap(
                colors=np.array(
                    [
                        [1.0, 0.0, 0.0],
                    ],
                    dtype=np.float32,
                )
            ),
            symbol="+",
            legend="corrected Pixels",
        )
        scatter.setSymbolSize(7)
        self._plot.setGraphTitle("Raw Pixels O  (blue) vs Corrected Pixels + (red)")
        self._plot.resetZoom()

    def _validateInputs(self) -> str:
        input_url = self._segmenter_url.text().strip()
        if not input_url:
            raise ValueError("No segmenter data URL to process.")
        return input_url



from __future__ import annotations

import shutil
from pathlib import Path
from typing import Any, List, Tuple

import h5py
import numpy as np
from ImageD11.grain import grain as Grain

from .models import SampleConfig, UnitCellParameters


def get_monitor_data(
    masterfile_path: str | Path, scan_number: str | int, monitor_name: str
):
    with h5py.File(masterfile_path, "r") as hin:
        instrument_group_1 = f"{scan_number}.1/instrument"
        if monitor_name in hin[instrument_group_1]:
            monitor_array = hin[instrument_group_1][monitor_name]["data"][:]
            return monitor_array

        instrument_group_2 = f"{scan_number}.2/instrument"
        if monitor_name in hin[instrument_group_2]:
            monitor_async_array = hin[instrument_group_2][monitor_name]["data"][:]
            t_sync_seq = hin[instrument_group_1]["epoch_trig/data"][:]
            t_async_seq = hin[instrument_group_2]["epoch"]["value"][:]
            if t_async_seq.shape[0] > monitor_async_array.shape[0]:
                t_async_seq = (t_async_seq[0:-1] + t_async_seq[1:]) / 2
            else:
                t_async_seq = t_async_seq

            monitor_array = np.interp(t_sync_seq, t_async_seq, monitor_async_array)
            return monitor_array

    raise KeyError(
        f""" Given monitor: {monitor_name} not found in the scan_number '{scan_number}''
                of given master file {masterfile_path}.
        """
    )


def get_monitor_scale_factor(
    masterfile_path: str | Path, scan_number: str | int, monitor_name: str
) -> np.ndarray:
    monitor_array = get_monitor_data(
        masterfile_path=masterfile_path,
        scan_number=scan_number,
        monitor_name=monitor_name,
    )
    return np.mean(monitor_array) / monitor_array


def get_optimal_frame_index(
    masterfile_path: str | Path,
    scan_number: str | int,
    detector: str,
    counter: str = "_roi1",
) -> int:
    with h5py.File(masterfile_path, "r") as h5file:
        detector_ctr = detector + counter
        scan_string = str(scan_number) + ".1"
        frame_idx = np.argmax(h5file[f"{scan_string}/measurement/{detector_ctr}"][:])
        return frame_idx


def get_monitor_scale_factor_for_frame_index(
    masterfile_path: str | Path,
    scan_number: str | int,
    detector: str,
    monitor_name: str,
    frame_idx: int | None,
    counter: str = "_roi1",
) -> float:
    monitor_scale_array = get_monitor_scale_factor(
        masterfile_path=masterfile_path,
        scan_number=scan_number,
        monitor_name=monitor_name,
    )

    if frame_idx is None:
        frame_idx = get_optimal_frame_index(
            masterfile_path=masterfile_path,
            scan_number=scan_number,
            detector=detector,
            counter=counter,
        )
    return monitor_scale_array[frame_idx]


def read_wavelength(geometry_par_file: str | Path) -> float:
    with open(geometry_par_file, "r") as f:
        for line in f:
            if line.startswith("wavelength"):
                try:
                    wavelength = float(line.strip().split()[1])
                    break
                except (IndexError, ValueError):
                    raise ValueError(
                        f"Could not found a valid wavelength field in {geometry_par_file}"
                    )
    return wavelength


def get_omega_slop(filepath: str | Path, scan_number: str | int, omega_motor: str):
    """Gets the mean variation of the omega motor, also called slop."""
    with h5py.File(filepath, "r") as hin:
        omega_angles = hin[f"{scan_number}.1/measurement"].get(omega_motor, None)
        if not isinstance(omega_angles, h5py.Dataset):
            raise TypeError(
                f'Could not find a dataset at {filepath}::{scan_number}.1/measurement/{omega_motor}". Got {omega_angles} instead'
            )
        omega_array = omega_angles[()]

    omegas_sorted = np.sort(omega_array)
    return np.round(np.diff(omegas_sorted).mean(), 3) / 2


def extract_sample_info(master_file: str) -> Tuple[str, str, str]:
    """
    Helper function to extract the dataroot, sample, dset_name, scan no from the
    masterfile path
    """
    path = Path(master_file)
    if len(path.parts) < 4:
        raise ValueError(
            "Expected path structure to be of the for `/dataroot/sample_name/dset_name/master_file.h5`"
        )

    # Extract dset name
    dset_folder_name = path.parent.name
    if "_" not in dset_folder_name:
        raise ValueError(
            f"Invalid dataset name: no underscore found in {dset_folder_name} to separate sample and dataset name.",
        )

    dataroot = path.parent.parent.parent.__str__()
    sample_name = path.parent.parent.name
    dset_name = dset_folder_name[len(sample_name + "_") :]
    return dataroot, sample_name, dset_name


def get_frame_image(
    file_path: str | Path,
    detector: str,
    scan_id: str,
    frame_idx: int | None = None,
    counter: str = "_roi1",
) -> np.ndarray:
    """
    extract a frame image from master file path
    """
    with h5py.File(file_path, "r") as h5file:
        if frame_idx is None:
            frame_idx = get_optimal_frame_index(
                masterfile_path=file_path,
                scan_number=scan_id[:-2],
                detector=detector,
                counter=counter,
            )
        return h5file[f"{scan_id}/measurement/{detector}"][frame_idx]


def find_omega_slop(cfg: SampleConfig):
    """Finds the mean variation of the omega motor, also called slop, in the scan file."""
    master_file = Path(cfg.master_file)
    if not master_file.exists():
        raise FileNotFoundError(
            f"""Could not find HDF5 master file at {master_file}."""
        )

    return get_omega_slop(
        filepath=master_file, scan_number=cfg.scan_number, omega_motor=cfg.omega_motor
    )


def copy_par_file(original_par_file: str, dest_folder: Path) -> Path:
    par_file = Path(original_par_file)

    if not par_file.exists():
        raise FileNotFoundError(f"Provided file '{par_file}' does not exist.")

    if par_file.suffix != ".par":
        raise ValueError(f"Provided file '{par_file}' is not a .par file")

    output_file = Path(dest_folder) / "par_folder" / par_file.name
    output_file.parent.mkdir(exist_ok=True)
    shutil.copy2(par_file, output_file)

    return output_file


def load_par_or_cif_file(filepath: str | Path) -> dict:
    """
    Read .par or .cif file, and return each line as key, value pair
    """

    read_parameters = {}
    with open(filepath, "r") as file:
        for line in file:
            parts = line.strip().split()
            if len(parts) != 2:
                continue
            key, value = parts
            read_parameters[key] = value

    return read_parameters


def save_par_file(filepath: str | Path, parameters: dict, mode="w"):
    """
    Saves a .par or .cif file
    """
    with open(filepath, mode) as f:
        for key, value in parameters.items():
            f.write(f"{key} {value}\n")


def read_lattice_cell_data(lattice_data_file_path: str | Path) -> UnitCellParameters:
    """
    Reads lattice cell data from a file and extracts the lattice parameters and space group.
    """
    return UnitCellParameters(**load_par_or_cif_file(lattice_data_file_path))


def save_geometry_and_lattice_file(
    file_path: str | Path, geom_dict: dict[str, Any], lattice_dict: dict[str, Any]
):
    with open(file_path, "w") as f:
        lattice_params = lattice_dict.get("lattice_parameters", [])
        if len(lattice_params) != 6:
            raise ValueError(
                f"Expected a list of 6 lattice params. Got {lattice_params}"
            )
        f.write(f"cell__a {lattice_params[0]}\n")
        f.write(f"cell__b {lattice_params[1]}\n")
        f.write(f"cell__c {lattice_params[2]}\n")
        f.write(f"cell_alpha {lattice_params[3]}\n")
        f.write(f"cell_beta {lattice_params[4]}\n")
        f.write(f"cell_gamma {lattice_params[5]}\n")

        if "lattice_space_group" in lattice_dict:
            f.write(
                f"cell_lattice_[P,A,B,C,I,F,R] {lattice_dict['lattice_space_group']}\n"
            )

    save_par_file(file_path, geom_dict, mode="a")


def save_grains_to_nexus(
    nexus_filename: str | Path,
    geometry_par_path: str | Path,
    lattice_file: str | Path,
    lattice_name: str,
    grains: List[Grain],
) -> None:
    """
    Save grain information (UBI matrices, lattice parameters) and
    the experiment geometry into a NeXus file.

    Parameters:
    - nexus_filename: Output NeXus file path
    - geometry_par_path: Path to the geometry_tdxrd.par file
    - lattice_file: Path to lattice parameters file
    - lattice_name: Name of the lattice type
    - grains (list of grain instances): List of grain objects with:
        - grain.ubi: (3,3) array
    """

    unit_cell_parameters = read_lattice_cell_data(lattice_data_file_path=lattice_file)
    lattice_parameters, space_group = (
        unit_cell_parameters.lattice_parameters,
        unit_cell_parameters.space_group,
    )

    with open(geometry_par_path, "r") as f:
        geometry_lines = [line.strip() for line in f]
    geometry_par = "\n".join(geometry_lines)

    num_grains = len(grains)
    ubi_matrices = np.zeros((num_grains, 3, 3), dtype=grains[0].ubi.dtype)
    translations = np.full((num_grains, 3), fill_value=np.nan)

    for i, grain in enumerate(grains):
        ubi_matrices[i] = grain.ubi
        if grain.translation is not None:
            translations[i] = grain.translation

    with h5py.File(nexus_filename, "w") as f:
        entry = f.create_group("entry")
        entry.attrs["NX_class"] = "NXentry"
        index_grains = entry.create_group("indexed_grains")
        index_grains.attrs["NX_class"] = "NXprocess"
        parameters_group = index_grains.create_group("parameters")
        parameters_group.attrs["NX_class"] = "NXcollection"
        parameters_group.create_dataset("geometry_par", data=geometry_par)
        parameters_group.create_dataset("lattice_name", data=lattice_name)
        parameters_group.create_dataset(
            "lattice_parameters", data=np.array(lattice_parameters)
        )
        if isinstance(space_group, int):
            parameters_group.create_dataset("space_group_number", data=space_group)
        else:
            parameters_group.create_dataset("space_group_symbol", data=space_group)

        grains_group = index_grains.create_group("grains")
        grains_group.attrs["NX_class"] = "NXdata"
        grains_group.create_dataset("UBI", data=ubi_matrices)
        grains_group.create_dataset("translation", data=translations)



from typing import Any, Iterable, Type

from silx.gui import qt


class TupleInputWidget(qt.QWidget):
    def __init__(
        self,
        placeHolderText: str = "Enter tuple, e.g. 0, 1, 2",
        Type: Type = int,
        parent: qt.QWidget | None = None,
    ):
        super().__init__(parent)

        self._tupleEdit = qt.QLineEdit()
        self._tupleEdit.setPlaceholderText(placeHolderText)

        layout = qt.QVBoxLayout(self)
        layout.addWidget(self._tupleEdit)
        layout.setContentsMargins(0, 0, 0, 0)

        self._type = Type
        self._tupleEdit.editingFinished.connect(self._validate)

    def _validate(self):
        try:
            _ = self.getValue()
            self._tupleEdit.setStyleSheet("")
        except Exception:
            self._tupleEdit.setStyleSheet("border: 1px solid red")

    def getValue(self) -> tuple:
        text = self._tupleEdit.text().strip()

        if not text:
            return tuple()

        try:
            return tuple(self._type(part) for part in text.split(",") if part.strip())
        except ValueError as e:
            raise ValueError(f"Input must be a tuple of {self._type.__name__}") from e

    def setValue(self, values: Iterable[Any]):
        self._tupleEdit.setText(",".join(str(value) for value in values))



from pathlib import Path
from typing import Optional, Union

import numpy as np
from ewokscore import Task
from pydantic import Field

from ..indexing import run_grid_indexing
from ..io import find_omega_slop
from ..models import GridIndexParameters, InputsWithOverwrite, SampleConfig
from ..nexus.grains import save_nexus_grain_process
from ..nexus.parameters import read_nx_collection_parameters
from ..nexus.utils import (
    check_throw_write_data_group_url,
    get_data_url_paths,
    get_entry_name,
)
from ..utils import generate_translations


class Inputs(InputsWithOverwrite):
    indexer_filtered_data_url: str = Field(
        description="Path to NXprocess group containing the filtered (be it by index, lattice or intensity) peaks"
    )
    grid_index_parameters: GridIndexParameters = Field(
        description="""dict of following parameters:
        - `NPKS`       (int)           : number of peaks
        - `DSTOL`      (float)         :  reciprocal distance tolerance
        - `RING1`      (Sequence[int]) : default (1, 0)
        - `RING2`      (Sequence[int]) : default (0,)
        - `NUL`        (bool)          : default True
        - `FITPOS`     (bool)          : default True
        - `tolangle`   (float)         : default 0.50
        - `toldist`    (float)         : default 100.0
        - `NPROC`      (int)           : default None, but ImageD11 will set it to maximum available cores on the server
        - `NTHREAD`    (int)           : default 1
        - `COSTOL`     (float)         : default None (from model) if default provided then here will set it to np.cos(np.radians(90 - 2 * omega_slop))
        - `OMEGAFLOAT` (float)         : default None (from model) if default provided then here will set it to omega_slop
            Note: omega_slop is related to wobble of the instrument,
        - `TOLSEQ` (Tuple[float, ...]) : default (0.02, 0.015, 0.01)
        - `SYMMETRY` (str)             : default cubic"""
    )
    analyse_folder: Union[str, Path, None] = None
    grid_abs_x_limit: int = Field(
        default=600,
        description="X limit of the 3D space to search translation positions of grains.",
    )
    grid_abs_y_limit: int = Field(
        default=600,
        description="Y limit of the 3D space to search translation positions of grains.",
    )
    grid_abs_z_limit: int = Field(
        default=200,
        description="Z limit of the 3D space to search translation positions of grains.",
    )
    grid_step: int = Field(
        default=100,
        description="Step size of the 3D space to search translation positions of grains.",
    )
    seed: Optional[int] = Field(
        default=42,
        description="Seed for translation shuffling. Pass `None` to disable translation shuffling",
    )
    sample_config: Optional[SampleConfig] = Field(
        default=None,
        description="Override the sample config. If not provided, it will be retrieved from the segmentation step",
    )


class GridIndexGrains(
    Task,
    input_model=Inputs,
    output_names=["grid_indexed_grain_data_url"],
):
    """
    From 3D peaks, finds grains' UBI  matrices and stores them in NeXus (.h5) format.

    ```{warning}
    This task is still experimental and may break or change without notice.
    ```

    Outputs:
    - `grid_indexed_grain_data_url` (str): Data path to 'NxProcess group data' grains that stores generated UBI

    """

    def run(self):
        inputs = Inputs(**self.get_input_values())
        nexus_file_path, indexer_filtered_data_group_path = get_data_url_paths(
            inputs.indexer_filtered_data_url
        )

        entry_name = get_entry_name(indexer_filtered_data_group_path)
        grid_indexed_ubi_data_group = f"{entry_name}/grid_indexed_grains"
        check_throw_write_data_group_url(
            overwrite=inputs.overwrite,
            filename=nexus_file_path,
            data_group_path=grid_indexed_ubi_data_group,
        )

        filename, process_group_name = get_data_url_paths(
            f"{nexus_file_path}::{entry_name}/segmented_3d_peaks"
        )
        sample_config = inputs.sample_config
        if sample_config is None:
            sample_config = SampleConfig(
                **read_nx_collection_parameters(
                    filename=filename,
                    process_group_name=process_group_name,
                    sub_folder="parameters/FolderFileSettings",
                )
            )

        omega_slop = find_omega_slop(sample_config)

        grid_index_parameters = inputs.grid_index_parameters
        if grid_index_parameters.COSTOL is None:
            grid_index_parameters.COSTOL = np.cos(np.radians(90 - 2 * omega_slop))
        if grid_index_parameters.OMEGAFLOAT is None:
            grid_index_parameters.OMEGAFLOAT = omega_slop

        analyse_folder = (
            Path(inputs.analyse_folder)
            if inputs.analyse_folder
            else Path(nexus_file_path).parent
        )

        grid_indexed_grains = run_grid_indexing(
            input_data_url=inputs.indexer_filtered_data_url,
            grid_index_parameters=grid_index_parameters,
            analyse_folder=analyse_folder,
            translations=generate_translations(
                max_x=inputs.grid_abs_x_limit,
                max_y=inputs.grid_abs_y_limit,
                max_z=inputs.grid_abs_z_limit,
                step=inputs.grid_step,
                seed=inputs.seed,
            ),
        )

        ubi_grains_group_name = save_nexus_grain_process(
            filename=nexus_file_path,
            entry_name=entry_name,
            process_name="grid_indexed_grains",
            config_settings={
                **grid_index_parameters.model_dump(exclude_none=True),
                "data_from": self.inputs.indexer_filtered_data_url,
                "grid_abs_x_limit": inputs.grid_abs_x_limit,
                "grid_abs_y_limit": inputs.grid_abs_y_limit,
                "grid_abs_z_limit": inputs.grid_abs_z_limit,
                "grid_step": inputs.grid_step,
            },
            grains=grid_indexed_grains,
            overwrite=inputs.overwrite,
        )
        self.outputs.grid_indexed_grain_data_url = ubi_grains_group_name
